# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

try:
    from elasticsearch.connection.base import Connection
except ImportError:
    from elastic_transport._models import NodeConfig
    from elastic_transport._node._base import BaseNode as Connection

from conftest import ES_VERSION, ES_SETTINGS


HOST = {"scheme": "http", "host": ES_SETTINGS["host"], "port": int(ES_SETTINGS["port"])}

IS_V8 = ES_VERSION >= (8,)
SKIP_IF_V7 = pytest.mark.skipif(not IS_V8, reason="Skipping v8 tests.")
SKIP_IF_V8 = pytest.mark.skipif(IS_V8, reason="Skipping v7 tests.")


def test_connection_default():
    if IS_V8:
        conn = Connection(NodeConfig(**HOST))
    else:
        conn = Connection(**HOST)

    assert conn._nr_host_port == ("localhost", ES_SETTINGS["port"])


@SKIP_IF_V7
def test_connection_config():
    conn = Connection(NodeConfig(scheme="http", host="foo", port=8888))
    assert conn._nr_host_port == ("foo", "8888")


@SKIP_IF_V8
def test_connection_host_arg():
    conn = Connection("the_host")
    assert conn._nr_host_port == ("the_host", "9200")


@SKIP_IF_V8
def test_connection_args():
    conn = Connection("the_host", 9999)
    assert conn._nr_host_port == ("the_host", "9999")


@SKIP_IF_V8
def test_connection_kwargs():
    conn = Connection(host="foo", port=8888)
    assert conn._nr_host_port == ("foo", "8888")




# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from io import BytesIO

import asyncpg
import pytest
from testing_support.db_settings import postgresql_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = postgresql_settings()[0]


PG_PREFIX = "Datastore/operation/Postgres/"
ASYNCPG_VERSION = tuple(int(x) for x in getattr(asyncpg, "__version__", "0.0").split(".")[:2])

if ASYNCPG_VERSION < (0, 11):
    CONNECT_METRICS = ()
else:
    CONNECT_METRICS = ((PG_PREFIX + "connect", 1),)


@pytest.fixture
def conn(event_loop):
    conn = event_loop.run_until_complete(
        asyncpg.connect(
            user=DB_SETTINGS["user"],
            password=DB_SETTINGS["password"],
            database=DB_SETTINGS["name"],
            host=DB_SETTINGS["host"],
            port=DB_SETTINGS["port"],
        )
    )
    yield conn
    event_loop.run_until_complete(conn.close())


@validate_transaction_metrics(
    "test_single",
    background_task=True,
    scoped_metrics=((PG_PREFIX + "select", 1),),
    rollup_metrics=(("Datastore/all", 1),),
)
@validate_tt_collector_json(datastore_params={"port_path_or_id": str(DB_SETTINGS["port"])})
@background_task(name="test_single")
@pytest.mark.parametrize("method", ("execute",))
def test_single(event_loop, method, conn):
    _method = getattr(conn, method)
    event_loop.run_until_complete(_method("""SELECT 0"""))


@validate_transaction_metrics(
    "test_prepared_single",
    background_task=True,
    scoped_metrics=(
        (PG_PREFIX + "prepare", 1),
        (PG_PREFIX + "select", 1),
    ),
    rollup_metrics=(("Datastore/all", 2),),
)
@background_task(name="test_prepared_single")
@pytest.mark.parametrize("method", ("fetch", "fetchrow", "fetchval"))
def test_prepared_single(event_loop, method, conn):
    _method = getattr(conn, method)
    event_loop.run_until_complete(_method("""SELECT 0"""))


@validate_transaction_metrics(
    "test_prepare",
    background_task=True,
    scoped_metrics=((PG_PREFIX + "prepare", 1),),
    rollup_metrics=(("Datastore/all", 1),),
)
@background_task(name="test_prepare")
def test_prepare(event_loop, conn):
    event_loop.run_until_complete(conn.prepare("""SELECT 0"""))


@pytest.fixture
def table(event_loop, conn):
    table_name = "table_%d" % os.getpid()

    event_loop.run_until_complete(conn.execute("""create table %s (a integer, b real, c text)""" % table_name))

    return table_name


@pytest.mark.skipif(ASYNCPG_VERSION < (0, 11), reason="Copy wasn't implemented before 0.11")
@validate_transaction_metrics(
    "test_copy",
    background_task=True,
    scoped_metrics=(
        (PG_PREFIX + "prepare", 1),
        (PG_PREFIX + "copy", 3),
    ),
    rollup_metrics=(("Datastore/all", 4),),
)
@background_task(name="test_copy")
def test_copy(event_loop, table, conn):
    async def amain():
        await conn.copy_records_to_table(table, records=[(1, 2, "3"), (4, 5, "6")])
        await conn.copy_from_table(table, output=BytesIO())

        # Causes a prepare and copy statement to be executed
        # 2 statements
        await conn.copy_from_query("""SELECT 0""", output=BytesIO())

    event_loop.run_until_complete(amain())


@validate_transaction_metrics(
    "test_select_many",
    background_task=True,
    scoped_metrics=(
        (PG_PREFIX + "prepare", 1),
        (PG_PREFIX + "select", 1),
    ),
    rollup_metrics=(("Datastore/all", 2),),
)
@background_task(name="test_select_many")
def test_select_many(event_loop, conn):
    event_loop.run_until_complete(conn.executemany("""SELECT $1::int""", ((1,), (2,))))


@validate_transaction_metrics(
    "test_transaction",
    background_task=True,
    scoped_metrics=(
        (PG_PREFIX + "begin", 1),
        (PG_PREFIX + "select", 1),
        (PG_PREFIX + "commit", 1),
    ),
    rollup_metrics=(("Datastore/all", 3),),
)
@background_task(name="test_transaction")
def test_transaction(event_loop, conn):
    async def amain():
        async with conn.transaction():
            await conn.execute("""SELECT 0""")

    event_loop.run_until_complete(amain())


@validate_transaction_metrics(
    "test_cursor",
    background_task=True,
    scoped_metrics=(
        (PG_PREFIX + "begin", 1),
        (PG_PREFIX + "prepare", 2),
        (PG_PREFIX + "select", 3),
        (PG_PREFIX + "commit", 1),
    ),
    rollup_metrics=(("Datastore/all", 7),),
)
@background_task(name="test_cursor")
def test_cursor(event_loop, conn):
    async def amain():
        async with conn.transaction():
            async for record in conn.cursor("SELECT generate_series(0, 0)", prefetch=1):
                pass

            await conn.cursor("SELECT 0")

    event_loop.run_until_complete(amain())


@pytest.mark.skipif(
    ASYNCPG_VERSION < (0, 11),
    reason="This is testing connect behavior which is only captured on newer asyncpg versions",
)
@validate_transaction_metrics(
    "test_unix_socket_connect",
    background_task=True,
    rollup_metrics=[
        (
            "Datastore/instance/Postgres/" + instance_hostname("localhost") + "//.s.PGSQL.THIS_FILE_BETTER_NOT_EXIST",
            1,
        )
    ],
)
@background_task(name="test_unix_socket_connect")
def test_unix_socket_connect(event_loop):
    with pytest.raises(OSError):
        event_loop.run_until_complete(asyncpg.connect("postgres://?host=/.s.PGSQL.THIS_FILE_BETTER_NOT_EXIST"))


@pytest.mark.skipif(
    ASYNCPG_VERSION < (0, 11),
    reason="This is testing connect behavior which is only captured on newer asyncpg versions",
)
@validate_transaction_metrics(
    "test_pool_acquire",
    background_task=True,
    scoped_metrics=((PG_PREFIX + "connect", 2),),
)
@background_task(name="test_pool_acquire")
def test_pool_acquire(event_loop):
    async def amain():
        pool = await asyncpg.create_pool(
            user=DB_SETTINGS["user"],
            password=DB_SETTINGS["password"],
            database=DB_SETTINGS["name"],
            host=DB_SETTINGS["host"],
            port=DB_SETTINGS["port"],
            min_size=1,
        )

        try:
            async with pool.acquire():
                async with pool.acquire():
                    pass

        finally:
            await pool.close()

    event_loop.run_until_complete(amain())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest
import flask.views

from _test_views import app

from conftest import async_handler_support

# Async view support added in flask v2
if async_handler_support:
    class TestAsyncView(flask.views.View):
        async def dispatch_request(self):
            return "ASYNC VIEW RESPONSE"

    class TestAsyncMethodView(flask.views.MethodView):
        async def get(self):
            return "ASYNC METHODVIEW GET RESPONSE"

    app.add_url_rule("/async_view", view_func=TestAsyncView.as_view("test_async_view"))
    app.add_url_rule(
        "/async_methodview",
        view_func=TestAsyncMethodView.as_view("test_async_methodview"),
    )

_test_application = webtest.TestApp(app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import socket
import threading
from urllib.request import HTTPError, urlopen

import pytest
import uvicorn
from testing_support.fixtures import (
    override_application_settings,
    raise_background_exceptions,
    wait_for_background_threads,
)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.sample_asgi_applications import (
    AppWithCall,
    AppWithCallRaw,
    simple_app_v2_raw,
)
from uvicorn.config import Config
from uvicorn.main import Server

from newrelic.common.object_names import callable_name

UVICORN_VERSION = tuple(int(v) for v in uvicorn.__version__.split(".")[:2])


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    return port


@pytest.fixture(
    params=(
        simple_app_v2_raw,
        pytest.param(
            AppWithCallRaw(),
            marks=pytest.mark.skipif(UVICORN_VERSION < (0, 6), reason="ASGI3 unsupported"),
        ),
        pytest.param(
            AppWithCall(),
            marks=pytest.mark.skipif(UVICORN_VERSION < (0, 6), reason="ASGI3 unsupported"),
        ),
    ),
    ids=("raw", "class_with_call", "class_with_call_double_wrapped"),
)
def app(request):
    return request.param


@pytest.fixture
def port(app):
    port = get_open_port()

    loops = []
    ready = threading.Event()

    def server_run():
        def on_tick_sync():
            if not ready.is_set():
                loops.append(asyncio.get_event_loop())
                ready.set()

        async def on_tick():
            on_tick_sync()

        config = Config(app, host="127.0.0.1", port=port, loop="asyncio")
        config.callback_notify = on_tick
        config.log_config = {"version": 1}
        config.disable_lifespan = True
        config.logger = logging.getLogger("uvicorn")
        server = Server(config=config)
        server.install_signal_handlers = lambda *args, **kwargs: None
        try:
            server.started.set = on_tick_sync
        except Exception:
            pass
        server.run()

    thread = threading.Thread(target=server_run, daemon=True)
    thread.start()
    assert ready.wait(timeout=10)
    yield port
    _ = [loop.stop() for loop in loops]  # Stop all loops
    thread.join(timeout=1)

    if thread.is_alive():
        raise RuntimeError("Thread failed to exit in time.")


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_uvicorn_200(port, app):
    @validate_transaction_metrics(callable_name(app))
    @raise_background_exceptions()
    @wait_for_background_threads()
    def response():
        return urlopen("http://localhost:%d" % port)

    assert response().status == 200


@override_application_settings({"transaction_name.naming_scheme": "framework"})
@validate_transaction_errors(["builtins:ValueError"])
def test_uvicorn_500(port, app):
    @validate_transaction_metrics(callable_name(app))
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        try:
            urlopen("http://localhost:%d/exc" % port)
        except HTTPError:
            pass

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import (register_database_client,
        enable_datastore_instance_feature)
from newrelic.api.transaction import current_transaction
from newrelic.core.database_node import DatabaseNode
from newrelic.core.stats_engine import StatsEngine

FIXTURE = os.path.join(os.curdir,
        'fixtures', 'datastores', 'datastore_instances.json')

_parameters_list = ['name', 'system_hostname', 'db_hostname',
        'product', 'port', 'unix_socket', 'database_path',
        'expected_instance_metric']

_parameters = ','.join(_parameters_list)


def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)


def _parametrize_test(test):
    return tuple([test.get(f, None if f != 'db_hostname' else 'localhost')
            for f in _parameters_list])


_datastore_tests = [_parametrize_test(t) for t in _load_tests()]


@pytest.mark.parametrize(_parameters, _datastore_tests)
@background_task()
def test_datastore_instance(name, system_hostname, db_hostname,
        product, port, unix_socket, database_path,
        expected_instance_metric, monkeypatch):

    monkeypatch.setattr('newrelic.common.system_info.gethostname',
            lambda: system_hostname)

    class FakeModule():
        pass

    register_database_client(FakeModule, product)
    enable_datastore_instance_feature(FakeModule)

    port_path_or_id = port or database_path or unix_socket

    node = DatabaseNode(dbapi2_module=FakeModule,
            sql='',
            children=[],
            start_time=0,
            end_time=1,
            duration=1,
            exclusive=1,
            stack_trace=None,
            sql_format='obfuscated',
            connect_params=None,
            cursor_params=None,
            sql_parameters=None,
            execute_params=None,
            host=db_hostname,
            port_path_or_id=port_path_or_id,
            database_name=database_path,
            guid=None,
            agent_attributes={},
            user_attributes={},
    )

    empty_stats = StatsEngine()
    transaction = current_transaction()
    unscoped_scope = ''

    # Check 'Datastore/instance' metric to confirm that:
    #   1. metric name is reported correctly
    #   2. metric scope is always unscoped

    for metric in node.time_metrics(empty_stats, transaction, None):
        if metric.name.startswith("Datastore/instance/"):
            assert metric.name == expected_instance_metric, name
            assert metric.scope == unscoped_scope



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.object_names import callable_name
from newrelic.packages import six


def test_serialization_metrics(skip_if_not_serializing, topic, send_producer_message):
    txn_name = "test_serialization:test_serialization_metrics.<locals>.test" if six.PY3 else "test_serialization:test"

    _metrics = [
        ("MessageBroker/Kafka/Topic/Named/%s/Serialization/Value" % topic, 1),
        ("MessageBroker/Kafka/Topic/Named/%s/Serialization/Key" % topic, 1),
    ]

    @validate_transaction_metrics(
        txn_name,
        scoped_metrics=_metrics,
        rollup_metrics=_metrics,
        background_task=True,
    )
    @background_task()
    def test():
        send_producer_message()

    test()


def test_deserialization_metrics(skip_if_not_serializing, topic, get_consumer_record):
    _metrics = [
        ("Message/Kafka/Topic/Named/%s/Deserialization/Value" % topic, 1),
        ("Message/Kafka/Topic/Named/%s/Deserialization/Key" % topic, 1),
    ]

    @validate_transaction_metrics(
        "Named/%s" % topic,
        group="Message/Kafka/Topic",
        scoped_metrics=_metrics,
        rollup_metrics=_metrics,
        background_task=True,
    )
    def test():
        get_consumer_record()

    test()


@pytest.mark.parametrize(
    "key,value,error",
    (
        (object(), "A", "KeySerializationError"),
        ("A", object(), "ValueSerializationError"),
    ),
)
def test_serialization_errors(skip_if_not_serializing, topic, producer, key, value, error):
    import confluent_kafka.error

    error_cls = getattr(confluent_kafka.error, error)

    @validate_transaction_errors([callable_name(error_cls)])
    @background_task()
    def test():
        with pytest.raises(error_cls):
            producer.produce(topic=topic, key=key, value=value)

    test()


@pytest.mark.parametrize(
    "key,value,error",
    (
        ("%", "{}", "KeyDeserializationError"),
        ("{}", "%", "ValueDeserializationError"),
    ),
)
def test_deserialization_errors(skip_if_not_serializing, monkeypatch, topic, producer, consumer, key, value, error):
    import confluent_kafka.error

    error_cls = getattr(confluent_kafka.error, error)

    # Remove serializers to cause intentional issues
    monkeypatch.setattr(producer, "_value_serializer", None)
    monkeypatch.setattr(producer, "_key_serializer", None)

    producer.produce(topic=topic, key=key, value=value)
    producer.flush()

    @validate_transaction_errors([callable_name(error_cls)])
    @background_task()
    def test():
        with pytest.raises(error_cls):
            timeout = 10
            attempts = 0
            while attempts < timeout:
                if not consumer.poll(0.5):
                    attempts += 1
                    continue

    test()


@pytest.fixture
def send_producer_message(topic, producer):
    def _test():
        producer.produce(topic, value={"foo": 1})
        producer.flush()

    return _test


@pytest.fixture()
def get_consumer_record(topic, send_producer_message, consumer):
    def _test():
        send_producer_message()

        record_count = 0

        timeout = 10
        attempts = 0
        record = None
        while not record and attempts < timeout:
            record = consumer.poll(0.5)
            if not record:
                attempts += 1
                continue
            assert not record.error()

            assert record.value() == {"foo": 1}
            record_count += 1
        consumer.poll(0.5)  # Exit the transaction.

        assert record_count == 1, "Incorrect count of records consumed: %d. Expected 1." % record_count

    return _test



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from flask import Flask
from flask import Blueprint
from werkzeug.routing import Rule

from conftest import is_flask_v2 as nested_blueprint_support

# Blueprints are only available in 0.7.0 onwards.

blueprint = Blueprint('blueprint', __name__)

application = Flask(__name__)

@blueprint.route('/index')
def index_page():
    return 'BLUEPRINT INDEX RESPONSE'

@blueprint.endpoint('endpoint')
def endpoint_page():
    return 'BLUEPRINT ENDPOINT RESPONSE'

@blueprint.before_app_first_request
def before_app_first_request():
    pass

@blueprint.before_request
def before_request():
    pass

@blueprint.before_app_request
def before_app_request():
    pass

@blueprint.after_request
def after_request(response):
    return response

@blueprint.after_app_request
def after_app_request(response):
    return response

@blueprint.teardown_request
def teardown_request(exc):
    pass

@blueprint.teardown_app_request
def teardown_app_request(exc):
    pass

# Support for nested blueprints was added in Flask 2.0
if nested_blueprint_support:
    parent = Blueprint('parent', __name__, url_prefix='/parent')
    child = Blueprint('child', __name__, url_prefix='/child')

    parent.register_blueprint(child)

    @child.route('/nested')
    def nested_page():
        return 'PARENT NESTED RESPONSE'
    application.register_blueprint(parent)

application.register_blueprint(blueprint)


application.url_map.add(Rule('/endpoint', endpoint='endpoint'))

_test_application = webtest.TestApp(application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper, function_wrapper
from newrelic.hooks.datastore_redis import (
    _redis_client_methods,
    _redis_multipart_commands,
    _redis_operation_re,
)
from newrelic.common.package_version_utils import get_package_version_tuple


def _conn_attrs_to_dict(connection):
    host = getattr(connection, "host", None)
    port = getattr(connection, "port", None)
    if not host and not port and hasattr(connection, "_address"):
        host, port = connection._address
    return {
        "host": host,
        "port": port,
        "path": getattr(connection, "path", None),
        "db": getattr(connection, "db", getattr(connection, "_db", None)),
    }


def _instance_info(kwargs):
    host = kwargs.get("host") or "localhost"
    port_path_or_id = str(kwargs.get("port") or kwargs.get("path", 6379))
    db = str(kwargs.get("db") or 0)

    return (host, port_path_or_id, db)


def _wrap_AioRedis_method_wrapper(module, instance_class_name, operation):

    @function_wrapper
    async def _nr_wrapper_AioRedis_async_method_(wrapped, instance, args, kwargs):
        transaction = current_transaction()
        if transaction is None:
            return await wrapped(*args, **kwargs)

        with DatastoreTrace(product="Redis", target=None, operation=operation):
            return await wrapped(*args, **kwargs)
    
    def _nr_wrapper_AioRedis_method_(wrapped, instance, args, kwargs):
        # Check for transaction and return early if found.
        # Method will return synchronously without executing,
        # it will be added to the command stack and run later.
        aioredis_version = get_package_version_tuple("aioredis")
        if aioredis_version and aioredis_version < (2,):
            # AioRedis v1 uses a RedisBuffer instead of a real connection for queueing up pipeline commands
            from aioredis.commands.transaction import _RedisBuffer
            if isinstance(instance._pool_or_conn, _RedisBuffer):
                # Method will return synchronously without executing,
                # it will be added to the command stack and run later.
                return wrapped(*args, **kwargs)
        else:
            # AioRedis v2 uses a Pipeline object for a client and internally queues up pipeline commands
            if aioredis_version:
                from aioredis.client import Pipeline
            else:
                from redis.asyncio.client import Pipeline
            if isinstance(instance, Pipeline):
                return wrapped(*args, **kwargs)

        # Method should be run when awaited, therefore we wrap in an async wrapper.
        return _nr_wrapper_AioRedis_async_method_(wrapped)(*args, **kwargs)


    name = "%s.%s" % (instance_class_name, operation)
    wrap_function_wrapper(module, name, _nr_wrapper_AioRedis_method_)


async def wrap_Connection_send_command(wrapped, instance, args, kwargs):
    transaction = current_transaction()
    if not transaction:
        return await wrapped(*args, **kwargs)

    host, port_path_or_id, db = (None, None, None)

    try:
        dt = transaction.settings.datastore_tracer
        if dt.instance_reporting.enabled or dt.database_name_reporting.enabled:
            conn_kwargs = _conn_attrs_to_dict(instance)
            host, port_path_or_id, db = _instance_info(conn_kwargs)
    except Exception:
        pass

    # Older Redis clients would when sending multi part commands pass
    # them in as separate arguments to send_command(). Need to therefore
    # detect those and grab the next argument from the set of arguments.

    operation = args[0].strip().lower()

    # If it's not a multi part command, there's no need to trace it, so
    # we can return early.

    if operation.split()[0] not in _redis_multipart_commands:        # Set the datastore info on the DatastoreTrace containing this function call.
        trace = current_trace()

        # Find DatastoreTrace no matter how many other traces are inbetween
        while trace is not None and not isinstance(trace, DatastoreTrace):
            trace = getattr(trace, "parent", None)

        if trace is not None:
            trace.host = host
            trace.port_path_or_id = port_path_or_id
            trace.database_name = db

        return await wrapped(*args, **kwargs)

    # Convert multi args to single arg string

    if operation in _redis_multipart_commands and len(args) > 1:
        operation = "%s %s" % (operation, args[1].strip().lower())

    operation = _redis_operation_re.sub("_", operation)

    with DatastoreTrace(
        product="Redis", target=None, operation=operation, host=host, port_path_or_id=port_path_or_id, database_name=db
    ):
        return await wrapped(*args, **kwargs)


def wrap_RedisConnection_execute(wrapped, instance, args, kwargs):
    # RedisConnection in aioredis v1 returns a future instead of using coroutines
    transaction = current_transaction()
    if not transaction:
        return wrapped(*args, **kwargs)

    host, port_path_or_id, db = (None, None, None)

    try:
        dt = transaction.settings.datastore_tracer
        if dt.instance_reporting.enabled or dt.database_name_reporting.enabled:
            conn_kwargs = _conn_attrs_to_dict(instance)
            host, port_path_or_id, db = _instance_info(conn_kwargs)
    except Exception:
        pass

    # Older Redis clients would when sending multi part commands pass
    # them in as separate arguments to send_command(). Need to therefore
    # detect those and grab the next argument from the set of arguments.

    operation = args[0].strip().lower()

    # If it's not a multi part command, there's no need to trace it, so
    # we can return early.

    if operation.split()[0] not in _redis_multipart_commands:        # Set the datastore info on the DatastoreTrace containing this function call.
        trace = current_trace()

        # Find DatastoreTrace no matter how many other traces are inbetween
        while trace is not None and not isinstance(trace, DatastoreTrace):
            trace = getattr(trace, "parent", None)

        if trace is not None:
            trace.host = host
            trace.port_path_or_id = port_path_or_id
            trace.database_name = db

        return wrapped(*args, **kwargs)

    # Convert multi args to single arg string

    if operation in _redis_multipart_commands and len(args) > 1:
        operation = "%s %s" % (operation, args[1].strip().lower())

    operation = _redis_operation_re.sub("_", operation)

    with DatastoreTrace(
        product="Redis", target=None, operation=operation, host=host, port_path_or_id=port_path_or_id, database_name=db
    ):
        return wrapped(*args, **kwargs)


def instrument_aioredis_client(module):
    # StrictRedis is just an alias of Redis, no need to wrap it as well.
    if hasattr(module, "Redis"):
        class_ = getattr(module, "Redis")
        for operation in _redis_client_methods:
            if hasattr(class_, operation):
                _wrap_AioRedis_method_wrapper(module, "Redis", operation)


def instrument_aioredis_connection(module):
    if hasattr(module, "Connection"):
        if hasattr(module.Connection, "send_command"):
            wrap_function_wrapper(module, "Connection.send_command", wrap_Connection_send_command)

    if hasattr(module, "RedisConnection"):
        if hasattr(module.RedisConnection, "execute"):
            wrap_function_wrapper(module, "RedisConnection.execute", wrap_RedisConnection_execute)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.encoding_utils import unpack_field
from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.common.system_info import LOCALHOST_EQUIVALENTS
from newrelic.core.database_utils import SQLConnections
from newrelic.packages import six


def _lookup_string_table(name, string_table, default=None):
    try:
        index = int(name.lstrip("`"))
        return string_table[index]
    except ValueError:
        return default


def validate_tt_collector_json(
    required_params=None,
    forgone_params=None,
    should_exist=True,
    datastore_params=None,
    datastore_forgone_params=None,
    message_broker_params=None,
    message_broker_forgone_params=None,
    exclude_request_uri=False,
):
    """make assertions based off the cross-agent spec on transaction traces"""
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    datastore_params = datastore_params or {}
    datastore_forgone_params = datastore_forgone_params or {}
    message_broker_params = message_broker_params or {}
    message_broker_forgone_params = message_broker_forgone_params or []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        traces_recorded = []

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        def _validate_tt_collector_json(wrapped, instance, args, kwargs):

            result = wrapped(*args, **kwargs)

            # Now that transaction has been recorded, generate
            # a transaction trace

            connections = SQLConnections()
            trace_data = instance.transaction_trace_data(connections)
            traces_recorded.append(trace_data)

            return result

        def _validate_trace(trace):
            assert isinstance(trace[0], float)  # absolute start time (ms)
            assert isinstance(trace[1], float)  # duration (ms)
            assert trace[0] > 0  # absolute time (ms)
            assert isinstance(trace[2], six.string_types)  # transaction name
            if trace[2].startswith("WebTransaction"):
                if exclude_request_uri:
                    assert trace[3] is None  # request url
                else:
                    assert isinstance(trace[3], six.string_types)
                    # query parameters should not be captured
                    assert "?" not in trace[3]

            # trace details -- python agent always uses condensed trace array

            trace_details, string_table = unpack_field(trace[4])
            assert len(trace_details) == 5
            assert isinstance(trace_details[0], float)  # start time (ms)

            # the next two items should be empty dicts, old parameters stuff,
            # placeholders for now

            assert isinstance(trace_details[1], dict)
            assert len(trace_details[1]) == 0
            assert isinstance(trace_details[2], dict)
            assert len(trace_details[2]) == 0

            # root node in slot 3

            root_node = trace_details[3]
            assert isinstance(root_node[0], float)  # entry timestamp
            assert isinstance(root_node[1], float)  # exit timestamp
            assert root_node[2] == "ROOT"
            assert isinstance(root_node[3], dict)
            assert len(root_node[3]) == 0  # spec shows empty (for root)
            children = root_node[4]
            assert isinstance(children, list)

            # there are two optional items at the end of trace segments,
            # class name that segment is in, and method name function is in;
            # Python agent does not use these (only Java does)

            # let's just test the first child
            trace_segment = children[0]
            assert isinstance(trace_segment[0], float)  # entry timestamp
            assert isinstance(trace_segment[1], float)  # exit timestamp
            assert isinstance(trace_segment[2], six.string_types)  # scope
            assert isinstance(trace_segment[3], dict)  # request params
            assert isinstance(trace_segment[4], list)  # children

            assert trace_segment[0] >= root_node[0]  # trace starts after root

            def _check_params_and_start_time(node):
                children = node[4]
                for child in children:
                    assert child[0] >= node[0]  # child started after parent
                    _check_params_and_start_time(child)

                params = node[3]
                assert isinstance(params, dict)

                # We should always report exclusive_duration_millis on a
                # segment. This allows us to override exclusive time
                # calculations on APM.
                assert "exclusive_duration_millis" in params
                assert isinstance(params["exclusive_duration_millis"], float)

                segment_name = _lookup_string_table(node[2], string_table, default=node[2])
                if segment_name.startswith("Datastore"):
                    for key in datastore_params:
                        assert key in params, key
                        assert params[key] == datastore_params[key]
                    for key in datastore_forgone_params:
                        assert key not in params, key

                    # if host is reported, it cannot be localhost
                    if "host" in params:
                        assert params["host"] not in LOCALHOST_EQUIVALENTS

                elif segment_name.startswith("MessageBroker"):
                    for key in message_broker_params:
                        assert key in params, key
                        assert params[key] == message_broker_params[key]
                    for key in message_broker_forgone_params:
                        assert key not in params, key

            _check_params_and_start_time(trace_segment)

            attributes = trace_details[4]

            assert "intrinsics" in attributes
            assert "userAttributes" in attributes
            assert "agentAttributes" in attributes

            assert isinstance(trace[5], six.string_types)  # GUID
            assert trace[6] is None  # reserved for future use
            assert trace[7] is False  # deprecated force persist flag

            # x-ray session ID

            assert trace[8] is None

            # Synthetics ID

            assert trace[9] is None or isinstance(trace[9], six.string_types)

            assert isinstance(string_table, list)
            for name in string_table:
                assert isinstance(name, six.string_types)  # metric name

        _new_wrapper = _validate_tt_collector_json(wrapped)
        val = _new_wrapper(*args, **kwargs)
        trace_data = traces_recorded.pop()
        trace = trace_data[0]  # 1st trace
        _validate_trace(trace)
        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import pytest
from test_application import is_graphql_2
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_span_events import validate_span_events

from newrelic.api.background_task import background_task


@pytest.fixture(scope="session")
def graphql_run_async():
    from graphql import __version__ as version
    from graphql import graphql

    major_version = int(version.split(".")[0])
    if major_version == 2:

        def graphql_run(*args, **kwargs):
            return graphql(*args, return_promise=True, **kwargs)

        return graphql_run
    else:
        return graphql


@dt_enabled
def test_query_and_mutation_async(app, graphql_run_async, is_graphql_2):
    from graphql import __version__ as version

    FRAMEWORK_METRICS = [
        ("Python/Framework/GraphQL/%s" % version, 1),
    ]
    _test_mutation_scoped_metrics = [
        ("GraphQL/resolve/GraphQL/storage", 1),
        ("GraphQL/resolve/GraphQL/storage_add", 1),
        ("GraphQL/operation/GraphQL/query/<anonymous>/storage", 1),
        ("GraphQL/operation/GraphQL/mutation/<anonymous>/storage_add", 1),
    ]
    _test_mutation_unscoped_metrics = [
        ("OtherTransaction/all", 1),
        ("GraphQL/all", 2),
        ("GraphQL/GraphQL/all", 2),
        ("GraphQL/allOther", 2),
        ("GraphQL/GraphQL/allOther", 2),
    ] + _test_mutation_scoped_metrics

    _expected_mutation_operation_attributes = {
        "graphql.operation.type": "mutation",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_mutation_resolver_attributes = {
        "graphql.field.name": "storage_add",
        "graphql.field.parentType": "Mutation",
        "graphql.field.path": "storage_add",
        "graphql.field.returnType": "[String]" if is_graphql_2 else "String",
    }
    _expected_query_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_query_resolver_attributes = {
        "graphql.field.name": "storage",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "storage",
        "graphql.field.returnType": "[String]",
    }

    @validate_transaction_metrics(
        "query/<anonymous>/storage",
        "GraphQL",
        scoped_metrics=_test_mutation_scoped_metrics,
        rollup_metrics=_test_mutation_unscoped_metrics + FRAMEWORK_METRICS,
        background_task=True,
    )
    @validate_span_events(exact_agents=_expected_mutation_operation_attributes)
    @validate_span_events(exact_agents=_expected_mutation_resolver_attributes)
    @validate_span_events(exact_agents=_expected_query_operation_attributes)
    @validate_span_events(exact_agents=_expected_query_resolver_attributes)
    @background_task()
    def _test():
        async def coro():
            response = await graphql_run_async(app, 'mutation { storage_add(string: "abc") }')
            assert not response.errors
            response = await graphql_run_async(app, "query { storage }")
            assert not response.errors

            # These are separate assertions because pypy stores 'abc' as a unicode string while other Python versions do not
            assert "storage" in str(response.data)
            assert "abc" in str(response.data)

        loop = asyncio.new_event_loop()
        loop.run_until_complete(coro())

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from minversion import pika_version_info


def basic_consume(channel, queue, callback, auto_ack=None):
    kwargs = {'queue': queue}
    if pika_version_info[0] < 1:
        kwargs['consumer_callback'] = callback
        if auto_ack is not None:
            kwargs['no_ack'] = not auto_ack
    else:
        kwargs['on_message_callback'] = callback
        if auto_ack is not None:
            kwargs['auto_ack'] = auto_ack

    return channel.basic_consume(**kwargs)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import aiohttp
import pytest
from testing_support.fixtures import (
    count_transactions,
    override_application_settings,
    override_expected_status_codes,
    override_generic_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.core.config import global_settings

version_info = tuple(int(_) for _ in aiohttp.__version__.split(".")[:2])


BASE_REQUIRED_ATTRS = ["request.headers.contentType", "request.method"]

# The agent should not record these attributes in events unless the settings
# explicitly say to do so
BASE_FORGONE_ATTRS = ["request.parameters.hello"]


@pytest.mark.parametrize("nr_enabled", [True, False])
@pytest.mark.parametrize(
    "method",
    [
        "GET",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
    ],
)
@pytest.mark.parametrize(
    "uri,metric_name,error,status",
    [
        ("/error?hello=world", "_target_application:error", "builtins:ValueError", 500),
        ("/non_500_error?hello=world", "_target_application:non_500_error", "aiohttp.web_exceptions:HTTPGone", 410),
        ("/raise_404?hello=world", "_target_application:raise_404", None, 404),
        ("/raise_403?hello=world", "_target_application:raise_403", "aiohttp.web_exceptions:HTTPForbidden", 403),
    ],
)
def test_error_exception(method, uri, metric_name, error, status, nr_enabled, aiohttp_app):
    async def fetch():
        resp = await aiohttp_app.client.request(method, uri, headers={"content-type": "text/plain"})
        assert resp.status == status

    required_attrs = list(BASE_REQUIRED_ATTRS)
    forgone_attrs = list(BASE_FORGONE_ATTRS)

    if nr_enabled:
        errors = []
        if error:
            errors.append(error)

        @validate_transaction_errors(
            errors=errors, expected_errors=["aiohttp.web_exceptions:HTTPForbidden"]
        )
        @validate_transaction_metrics(
            metric_name,
            scoped_metrics=[
                ("Function/%s" % metric_name, 1),
            ],
            rollup_metrics=[
                ("Function/%s" % metric_name, 1),
                ("Python/Framework/aiohttp/%s" % aiohttp.__version__, 1),
            ],
        )
        @validate_transaction_event_attributes(
            required_params={
                "agent": required_attrs,
                "user": [],
                "intrinsic": [],
            },
            forgone_params={
                "agent": forgone_attrs,
                "user": [],
                "intrinsic": [],
            },
            exact_attrs={
                "agent": {
                    "response.status": str(status),
                },
                "user": {},
                "intrinsic": {},
            },
        )
        @validate_code_level_metrics(*metric_name.split(":"))
        @override_ignore_status_codes([404])
        @override_expected_status_codes([403])
        def _test():
            aiohttp_app.loop.run_until_complete(fetch())

    else:
        settings = global_settings()

        @override_generic_settings(settings, {"enabled": False})
        def _test():
            aiohttp_app.loop.run_until_complete(fetch())

    _test()


@pytest.mark.parametrize("nr_enabled", [True, False])
@pytest.mark.parametrize(
    "method",
    [
        "GET",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
    ],
)
@pytest.mark.parametrize(
    "uri,metric_name",
    [
        ("/coro?hello=world", "_target_application:index"),
        ("/class?hello=world", "_target_application:HelloWorldView._respond"),
        ("/known_error?hello=world", "_target_application:KnownErrorView._respond"),
    ],
)
def test_simultaneous_requests(method, uri, metric_name, nr_enabled, aiohttp_app):
    async def fetch():
        resp = await aiohttp_app.client.request(method, uri, headers={"content-type": "text/plain"})
        assert resp.status == 200
        text = await resp.text()
        assert "Hello Aiohttp!" in text
        return resp

    async def multi_fetch(loop):
        coros = [fetch() for i in range(2)]

        try:
            combined = asyncio.gather(*coros)
        except TypeError:
            combined = asyncio.gather(*coros, loop=loop)

        responses = await combined
        return responses

    required_attrs = list(BASE_REQUIRED_ATTRS)
    extra_required = list(BASE_FORGONE_ATTRS)

    required_attrs.extend(extra_required)

    required_attrs.extend(["response.status", "response.headers.contentType"])

    if nr_enabled:
        transactions = []

        func_name = metric_name.replace(":", ".").split(".")
        namespace, func_name = ".".join(func_name[:-1]), func_name[-1]

        @override_application_settings({"attributes.include": ["request.*"]})
        @validate_transaction_metrics(
            metric_name,
            scoped_metrics=[
                ("Function/%s" % metric_name, 1),
            ],
            rollup_metrics=[
                ("Function/%s" % metric_name, 1),
                ("Python/Framework/aiohttp/%s" % aiohttp.__version__, 1),
            ],
        )
        @validate_transaction_event_attributes(
            required_params={
                "agent": required_attrs,
                "user": [],
                "intrinsic": [],
            },
            forgone_params={
                "agent": [],
                "user": [],
                "intrinsic": [],
            },
        )
        @validate_code_level_metrics(namespace, func_name)
        @count_transactions(transactions)
        def _test():
            aiohttp_app.loop.run_until_complete(multi_fetch(aiohttp_app.loop))
            assert len(transactions) == 2

    else:
        settings = global_settings()

        @override_generic_settings(settings, {"enabled": False})
        def _test():
            aiohttp_app.loop.run_until_complete(multi_fetch(aiohttp_app.loop))

    _test()


@pytest.mark.parametrize("nr_enabled", [True, False])
def test_system_response_creates_no_transaction(nr_enabled, aiohttp_app):
    async def fetch():
        resp = await aiohttp_app.client.request("GET", "/404")
        assert resp.status == 404
        return resp

    if nr_enabled:
        transactions = []

        @count_transactions(transactions)
        def _test():
            aiohttp_app.loop.run_until_complete(fetch())
            assert len(transactions) == 0

    else:
        settings = global_settings()

        @override_generic_settings(settings, {"enabled": False})
        def _test():
            aiohttp_app.loop.run_until_complete(fetch())

    _test()


def test_aborted_connection_creates_transaction(aiohttp_app):
    async def fetch():
        try:
            await aiohttp_app.client.request("GET", "/hang", timeout=0.1)
        except asyncio.TimeoutError:
            try:
                # Force the client to disconnect (while the server is hanging)
                await aiohttp_app.client.close()
            # In aiohttp 1.X, this can result in a CancelledError being raised
            except asyncio.CancelledError:
                pass
            await asyncio.sleep(0)
            return

        assert False, "Request did not time out"

    transactions = []

    @count_transactions(transactions)
    def _test():
        aiohttp_app.loop.run_until_complete(fetch())
        assert len(transactions) == 1

    _test()


def test_work_after_request_not_recorded(aiohttp_app):
    resp = aiohttp_app.loop.run_until_complete(aiohttp_app.client.request("GET", "/background"))
    assert resp.status == 200

    async def timeout():
        await asyncio.sleep(1)
        aiohttp_app.loop.stop()
        assert False

    task = aiohttp_app.loop.create_task(timeout())
    aiohttp_app.loop.run_forever()

    # Check that the timeout didn't fire
    assert not task.done()
    task.cancel()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import ssl
import tempfile
from collections import namedtuple

import pytest

from newrelic.common import certs, system_info
from newrelic.common.agent_http import DeveloperModeClient
from newrelic.common.encoding_utils import json_decode, serverless_payload_decode
from newrelic.common.utilization import CommonUtilization
from newrelic.core.agent_protocol import AgentProtocol, ServerlessModeProtocol
from newrelic.core.config import finalize_application_settings, global_settings
from newrelic.core.internal_metrics import InternalTraceContext
from newrelic.core.stats_engine import CustomMetrics
from newrelic.network.exceptions import (
    DiscardDataForRequest,
    ForceAgentDisconnect,
    ForceAgentRestart,
    NetworkInterfaceException,
    RetryDataForRequest,
)
from newrelic.packages import six

Request = namedtuple("Request", ("method", "path", "params", "headers", "payload"))


# Global constants used in tests
APP_NAME = "test_app"
IP_ADDRESS = AWS = AZURE = GCP = PCF = BOOT_ID = DOCKER = KUBERNETES = None
BROWSER_MONITORING_DEBUG = "debug"
BROWSER_MONITORING_LOADER = "loader"
CAPTURE_PARAMS = "capture_params"
DISPLAY_NAME = "display_name"
METADATA = {}
ENVIRONMENT = [["Agent Version", "test"]]
HIGH_SECURITY = True
HOST = "test_host"
LABELS = "labels"
LINKED_APPS = ["linked_app_1", "linked_app_2"]
MEMORY = 12000.0
PAYLOAD_APP_NAME = [APP_NAME] + LINKED_APPS
PAYLOAD_ID = ",".join(PAYLOAD_APP_NAME)
PID = 123
PROCESSOR_COUNT = 4
RECORD_SQL = "record_sql"
ANALYTIC_EVENT_DATA = 10000
SPAN_EVENT_DATA = 1000
CUSTOM_EVENT_DATA = 10000
ERROR_EVENT_DATA = 100


class HttpClientRecorder(DeveloperModeClient):
    SENT = []
    STATUS_CODE = None
    STATE = 0

    def send_request(
        self,
        method="POST",
        path="/agent_listener/invoke_raw_method",
        params=None,
        headers=None,
        payload=None,
    ):
        request = Request(method=method, path=path, params=params, headers=headers, payload=payload)
        self.SENT.append(request)
        if self.STATUS_CODE:
            return self.STATUS_CODE, b""

        return super(HttpClientRecorder, self).send_request(method, path, params, headers, payload)

    def __enter__(self):
        HttpClientRecorder.STATE += 1

    def __exit__(self, exc, value, tb):
        HttpClientRecorder.STATE -= 1

    def close_connection(self):
        HttpClientRecorder.STATE -= 1


class HttpClientException(DeveloperModeClient):
    def send_request(
        self,
        method="POST",
        path="/agent_listener/invoke_raw_method",
        params=None,
        headers=None,
        payload=None,
    ):
        raise NetworkInterfaceException


@pytest.fixture(autouse=True)
def clear_sent_values():
    yield
    HttpClientRecorder.SENT[:] = []
    HttpClientRecorder.STATUS_CODE = None
    HttpClientRecorder.STATE = 0


@pytest.fixture(autouse=True)
def override_utilization(monkeypatch):
    global AWS, AZURE, GCP, PCF, BOOT_ID, DOCKER, KUBERNETES
    AWS = {"id": "foo", "type": "bar", "zone": "baz"}
    AZURE = {"location": "foo", "name": "bar", "vmId": "baz", "vmSize": "boo"}
    GCP = {"id": 1, "machineType": "trmntr-t1000", "name": "arnold", "zone": "abc"}
    PCF = {"cf_instance_guid": "1", "cf_instance_ip": "7", "memory_limit": "0"}
    BOOT_ID = "cca356a7d72737f645a10c122ebbe906"
    DOCKER = {"id": "foobar"}
    KUBERNETES = {"kubernetes_service_host": "10.96.0.1"}

    @classmethod
    def detect(cls):
        name = cls.__name__
        output = None
        if name.startswith("BootId"):
            output = BOOT_ID
        elif name.startswith("AWS"):
            output = AWS
        elif name.startswith("Azure"):
            output = AZURE
        elif name.startswith("GCP"):
            output = GCP
        elif name.startswith("PCF"):
            output = PCF
        elif name.startswith("Docker"):
            output = DOCKER
        elif name.startswith("Kubernetes"):
            output = KUBERNETES
        else:
            assert False, "Unknown utilization class"

        if output is Exception:
            raise Exception
        return output

    monkeypatch.setattr(CommonUtilization, "detect", detect)


@pytest.fixture(autouse=True)
def override_system_info(monkeypatch):
    global IP_ADDRESS
    IP_ADDRESS = ["127.0.0.1"]
    monkeypatch.setattr(system_info, "gethostname", lambda *args, **kwargs: HOST)
    monkeypatch.setattr(system_info, "getips", lambda *args, **kwargs: IP_ADDRESS)
    monkeypatch.setattr(system_info, "logical_processor_count", lambda *args, **kwargs: PROCESSOR_COUNT)
    monkeypatch.setattr(system_info, "total_physical_memory", lambda *args, **kwargs: MEMORY)
    monkeypatch.setattr(os, "getpid", lambda *args, **kwargs: PID)


@pytest.mark.parametrize("status_code", (None, 202))
def test_send(status_code):
    HttpClientRecorder.STATUS_CODE = status_code
    settings = finalize_application_settings(
        {
            "request_headers_map": {"custom-header": u"value"},
            "agent_run_id": "RUN_TOKEN",
        }
    )
    protocol = AgentProtocol(settings, client_cls=HttpClientRecorder)
    response = protocol.send("metric_data", (1, 2, 3))
    assert response is None

    assert len(HttpClientRecorder.SENT) == 1
    request = HttpClientRecorder.SENT[0]

    assert request.method == "POST"
    assert request.path == "/agent_listener/invoke_raw_method"

    # Verify license key was there, but no way to validate the value
    request.params.pop("license_key")

    assert request.params == {
        "method": "metric_data",
        "marshal_format": "json",
        "protocol_version": AgentProtocol.VERSION,
        "run_id": "RUN_TOKEN",
    }

    assert request.headers == {
        "Content-Type": "application/json",
        "custom-header": u"value",
    }

    assert request.payload == b"[1,2,3]"

    # Verify call to finalize is None
    assert protocol.finalize() is None


@pytest.mark.parametrize(
    "status_code,expected_exc,log_level",
    (
        (400, DiscardDataForRequest, "WARNING"),
        (401, ForceAgentRestart, "ERROR"),
        (403, DiscardDataForRequest, "WARNING"),
        (404, DiscardDataForRequest, "WARNING"),
        (405, DiscardDataForRequest, "WARNING"),
        (407, DiscardDataForRequest, "WARNING"),
        (408, RetryDataForRequest, "INFO"),
        (409, ForceAgentRestart, "INFO"),
        (410, ForceAgentDisconnect, "CRITICAL"),
        (411, DiscardDataForRequest, "WARNING"),
        (413, DiscardDataForRequest, "WARNING"),
        (414, DiscardDataForRequest, "WARNING"),
        (415, DiscardDataForRequest, "WARNING"),
        (417, DiscardDataForRequest, "WARNING"),
        (429, RetryDataForRequest, "WARNING"),
        (431, DiscardDataForRequest, "WARNING"),
        (500, RetryDataForRequest, "WARNING"),
        (503, RetryDataForRequest, "WARNING"),
        (800, DiscardDataForRequest, "WARNING"),
    ),
)
def test_status_code_exceptions(status_code, expected_exc, log_level, caplog):
    caplog.set_level(logging.INFO)
    HttpClientRecorder.STATUS_CODE = status_code
    settings = finalize_application_settings(
        {
            "license_key": "123LICENSEKEY",
        }
    )
    protocol = AgentProtocol(settings, client_cls=HttpClientRecorder)

    internal_metrics = CustomMetrics()
    with InternalTraceContext(internal_metrics):
        with pytest.raises(expected_exc):
            protocol.send("analytic_event_data")

    internal_metrics = dict(internal_metrics.metrics())
    if status_code == 413:
        assert internal_metrics["Supportability/Python/Collector/MaxPayloadSizeLimit/analytic_event_data"] == [
            1,
            0,
            0,
            0,
            0,
            0,
        ]
    else:
        assert "Supportability/Python/Collector/MaxPayloadSizeLimit/analytic_event_data" not in internal_metrics

    assert len(HttpClientRecorder.SENT) == 1
    request = HttpClientRecorder.SENT[0]
    assert request.params["method"] == "analytic_event_data"

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == log_level
    message = caplog.records[0].getMessage()
    assert "123LICENSEKEY" not in message


def test_protocol_http_error_causes_retry():
    protocol = AgentProtocol(finalize_application_settings(), client_cls=HttpClientException)
    with pytest.raises(RetryDataForRequest):
        protocol.send("analytic_event_data")


def test_protocol_context_manager():
    protocol = AgentProtocol(finalize_application_settings(), client_cls=HttpClientRecorder)
    with protocol:
        assert HttpClientRecorder.STATE == 1

    assert HttpClientRecorder.STATE == 0


def test_close_connection():
    protocol = AgentProtocol(finalize_application_settings(), client_cls=HttpClientRecorder)
    protocol.close_connection()
    assert HttpClientRecorder.STATE == -1


def connect_payload_asserts(
    payload,
    with_aws=True,
    with_gcp=True,
    with_pcf=True,
    with_azure=True,
    with_docker=True,
    with_kubernetes=True,
):
    payload_data = payload[0]
    assert isinstance(payload_data["agent_version"], type(u""))
    assert payload_data["app_name"] == PAYLOAD_APP_NAME
    assert payload_data["display_host"] == DISPLAY_NAME
    assert payload_data["environment"] == ENVIRONMENT
    assert payload_data["metadata"] == METADATA
    assert payload_data["high_security"] == HIGH_SECURITY
    assert payload_data["host"] == HOST
    assert payload_data["identifier"] == PAYLOAD_ID
    assert payload_data["labels"] == LABELS
    assert payload_data["language"] == "python"
    assert payload_data["pid"] == PID
    assert len(payload_data["security_settings"]) == 2
    assert payload_data["security_settings"]["capture_params"] == CAPTURE_PARAMS
    assert payload_data["security_settings"]["transaction_tracer"] == {"record_sql": RECORD_SQL}
    assert len(payload_data["settings"]) == 2
    assert payload_data["settings"]["browser_monitoring.loader"] == (BROWSER_MONITORING_LOADER)
    assert payload_data["settings"]["browser_monitoring.debug"] == (BROWSER_MONITORING_DEBUG)

    utilization_len = 5

    assert "full_hostname" not in payload_data["utilization"]

    if IP_ADDRESS:
        assert payload_data["utilization"]["ip_address"] == IP_ADDRESS
        utilization_len += 1
    else:
        assert "ip_address" not in payload_data["utilization"]

    utilization_len = utilization_len + any([with_aws, with_pcf, with_gcp, with_azure, with_docker, with_kubernetes])
    assert len(payload_data["utilization"]) == utilization_len
    assert payload_data["utilization"]["hostname"] == HOST

    assert payload_data["utilization"]["logical_processors"] == PROCESSOR_COUNT
    assert payload_data["utilization"]["metadata_version"] == 5
    assert payload_data["utilization"]["total_ram_mib"] == MEMORY
    assert payload_data["utilization"]["boot_id"] == BOOT_ID

    # Faster Event Harvest
    harvest_limits = payload_data["event_harvest_config"]["harvest_limits"]
    assert harvest_limits["analytic_event_data"] == ANALYTIC_EVENT_DATA
    assert harvest_limits["span_event_data"] == SPAN_EVENT_DATA
    assert harvest_limits["custom_event_data"] == CUSTOM_EVENT_DATA
    assert harvest_limits["error_event_data"] == ERROR_EVENT_DATA

    vendors_len = 0

    if any([with_aws, with_pcf, with_gcp, with_azure]):
        vendors_len += 1

    if with_docker:
        vendors_len += 1

    if with_kubernetes:
        vendors_len += 1

    if vendors_len:
        assert len(payload_data["utilization"]["vendors"]) == vendors_len

        # check ordering
        if with_aws:
            assert payload_data["utilization"]["vendors"]["aws"] == AWS
        elif with_pcf:
            assert payload_data["utilization"]["vendors"]["pcf"] == PCF
        elif with_gcp:
            assert payload_data["utilization"]["vendors"]["gcp"] == GCP
        elif with_azure:
            assert payload_data["utilization"]["vendors"]["azure"] == AZURE

        if with_docker:
            assert payload_data["utilization"]["vendors"]["docker"] == DOCKER

        if with_kubernetes:
            assert payload_data["utilization"]["vendors"]["kubernetes"] == KUBERNETES
    else:
        assert "vendors" not in payload_data["utilization"]


@pytest.mark.parametrize(
    "with_aws,with_pcf,with_gcp,with_azure,with_docker,with_kubernetes,with_ip",
    [
        (False, False, False, False, False, False, False),
        (False, False, False, False, False, False, True),
        (True, False, False, False, True, True, True),
        (False, True, False, False, True, True, True),
        (False, False, True, False, True, True, True),
        (False, False, False, True, True, True, True),
        (True, False, False, False, False, False, True),
        (False, True, False, False, False, False, True),
        (False, False, True, False, False, False, True),
        (False, False, False, True, False, False, True),
        (True, True, True, True, True, True, True),
        (True, True, True, True, True, False, True),
        (True, True, True, True, False, True, True),
    ],
)
def test_connect(with_aws, with_pcf, with_gcp, with_azure, with_docker, with_kubernetes, with_ip):
    global AWS, AZURE, GCP, PCF, BOOT_ID, DOCKER, KUBERNETES, IP_ADDRESS
    if not with_aws:
        AWS = Exception
    if not with_pcf:
        PCF = Exception
    if not with_gcp:
        GCP = Exception
    if not with_azure:
        AZURE = Exception
    if not with_docker:
        DOCKER = Exception
    if not with_kubernetes:
        KUBERNETES = Exception
    if not with_ip:
        IP_ADDRESS = None
    settings = finalize_application_settings(
        {
            "browser_monitoring.loader": BROWSER_MONITORING_LOADER,
            "browser_monitoring.debug": BROWSER_MONITORING_DEBUG,
            "capture_params": CAPTURE_PARAMS,
            "process_host.display_name": DISPLAY_NAME,
            "transaction_tracer.record_sql": RECORD_SQL,
            "high_security": HIGH_SECURITY,
            "labels": LABELS,
            "utilization.detect_aws": with_aws,
            "utilization.detect_pcf": with_pcf,
            "utilization.detect_gcp": with_gcp,
            "utilization.detect_azure": with_azure,
            "utilization.detect_docker": with_docker,
            "utilization.detect_kubernetes": with_kubernetes,
            "event_harvest_config": {
                "harvest_limits": {
                    "analytic_event_data": ANALYTIC_EVENT_DATA,
                    "span_event_data": SPAN_EVENT_DATA,
                    "custom_event_data": CUSTOM_EVENT_DATA,
                    "error_event_data": ERROR_EVENT_DATA,
                }
            },
        }
    )
    protocol = AgentProtocol.connect(
        APP_NAME,
        LINKED_APPS,
        ENVIRONMENT,
        settings,
        client_cls=HttpClientRecorder,
    )

    # verify there are exactly 3 calls to HttpClientRecorder
    assert len(HttpClientRecorder.SENT) == 3

    # Verify preconnect call
    preconnect = HttpClientRecorder.SENT[0]
    assert preconnect.params["method"] == "preconnect"
    assert preconnect.payload == b"[]"

    # Verify connect call
    connect = HttpClientRecorder.SENT[1]
    assert connect.params["method"] == "connect"
    connect_payload = json_decode(connect.payload.decode("utf-8"))
    connect_payload_asserts(
        connect_payload,
        with_aws=with_aws,
        with_pcf=with_pcf,
        with_gcp=with_gcp,
        with_azure=with_azure,
        with_docker=with_docker,
        with_kubernetes=with_kubernetes,
    )

    # Verify agent_settings call is done with the finalized settings
    agent_settings = HttpClientRecorder.SENT[2]
    assert agent_settings.params["method"] == "agent_settings"
    agent_settings_payload = json_decode(agent_settings.payload.decode("utf-8"))
    assert len(agent_settings_payload) == 1
    agent_settings_payload = agent_settings_payload[0]

    # Finalized settings will have a non-None agent_run_id
    assert agent_settings_payload["agent_run_id"] is not None
    assert protocol.configuration.agent_run_id is not None

    # Verify that agent settings sent have converted null, containers, and
    # unserializable types to string
    assert agent_settings_payload["proxy_host"] == "None"
    assert agent_settings_payload["attributes.include"] == "[]"
    assert agent_settings_payload["feature_flag"] == str(set())
    assert isinstance(agent_settings_payload["attribute_filter"], six.string_types)

    # Verify that the connection is closed
    assert HttpClientRecorder.STATE == 0


def test_connect_metadata(monkeypatch):
    monkeypatch.setenv("NEW_RELIC_METADATA_FOOBAR", "foobar")
    monkeypatch.setenv("_NEW_RELIC_METADATA_WRONG", "wrong")
    AgentProtocol.connect(
        APP_NAME,
        LINKED_APPS,
        ENVIRONMENT,
        finalize_application_settings(),
        client_cls=HttpClientRecorder,
    )
    connect = HttpClientRecorder.SENT[1]
    assert connect.params["method"] == "connect"
    connect_payload = json_decode(connect.payload.decode("utf-8"))[0]
    assert connect_payload["metadata"] == {"NEW_RELIC_METADATA_FOOBAR": "foobar"}


def test_serverless_protocol_connect():
    settings = global_settings()
    protocol = ServerlessModeProtocol.connect(
        "testapp",
        ["foo"],
        [("Agent Version", "test")],
        settings,
        client_cls=HttpClientRecorder,
    )

    # No client calls should be made
    assert len(HttpClientRecorder.SENT) == 0

    # cross application tracing must be disabled
    assert not protocol.configuration.cross_application_tracer.enabled


def test_serverless_protocol_finalize(capsys):
    protocol = ServerlessModeProtocol(
        finalize_application_settings({"aws_lambda_metadata": {"foo": "bar", "agent_version": "x"}})
    )
    response = protocol.send("metric_data", (1, 2, 3))
    assert response is None

    payload = protocol.finalize()
    captured = capsys.readouterr()
    assert captured.out.rstrip("\n") == payload

    payload = json_decode(payload)
    assert payload[:2] == [1, "NR_LAMBDA_MONITORING"]

    data = serverless_payload_decode(payload[2])
    assert data["data"] == {"metric_data": [1, 2, 3]}

    assert data["metadata"]["foo"] == "bar"
    assert data["metadata"]["agent_version"] != "x"


def test_audit_logging():
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(b"*\n")

    settings = finalize_application_settings({"audit_log_file": f.name})
    protocol = AgentProtocol(settings, client_cls=HttpClientRecorder)
    protocol.send("preconnect")

    with open(f.name) as f:
        audit_log_contents = f.read()

    assert audit_log_contents.startswith("*\n")
    assert len(audit_log_contents) > 2


@pytest.mark.parametrize(
    "ca_bundle_path",
    (
        None,
        "custom",
    ),
)
def test_ca_bundle_path(monkeypatch, ca_bundle_path):
    # Pretend CA certificates are not available
    class DefaultVerifyPaths(object):
        cafile = None

        def __init__(self, *args, **kwargs):
            pass

    monkeypatch.setattr(ssl, "DefaultVerifyPaths", DefaultVerifyPaths)

    settings = finalize_application_settings({"ca_bundle_path": ca_bundle_path})
    protocol = AgentProtocol(settings)
    expected = ca_bundle_path or certs.where()
    assert protocol.client._connection_kwargs["ca_certs"] == expected


def test_max_payload_size_limit():
    settings = finalize_application_settings({"max_payload_size_in_bytes": 0, "port": -1})
    protocol = AgentProtocol(settings, host="localhost")
    with pytest.raises(DiscardDataForRequest):
        protocol.send("metric_data")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import uuid

import kafka
import pytest
from testing_support.db_settings import kafka_settings
from testing_support.fixtures import (  # noqa: F401, pylint: disable=W0611
    code_coverage_fixture,
    collector_agent_registration_fixture,
    collector_available_fixture,
)

from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import transient_function_wrapper

DB_SETTINGS = kafka_settings()[0]

BOOTSTRAP_SERVER = "%s:%s" % (DB_SETTINGS["host"], DB_SETTINGS["port"])
BROKER = [BOOTSTRAP_SERVER]

_coverage_source = [
    "newrelic.hooks.messagebroker_kafkapython",
]

code_coverage = code_coverage_fixture(source=_coverage_source)

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (messagebroker_kafkapython)",
    default_settings=_default_settings,
    linked_applications=["Python Agent Test (messagebroker_kafkapython)"],
)


@pytest.fixture(
    scope="session", params=["no_serializer", "serializer_function", "callable_object", "serializer_object"]
)
def client_type(request):
    return request.param


@pytest.fixture()
def skip_if_not_serializing(client_type):
    if client_type == "no_serializer":
        pytest.skip("Only serializing clients supported.")


@pytest.fixture(scope="function")
def producer(client_type, json_serializer, json_callable_serializer):
    if client_type == "no_serializer":
        producer = kafka.KafkaProducer(bootstrap_servers=BROKER)
    elif client_type == "serializer_function":
        producer = kafka.KafkaProducer(
            bootstrap_servers=BROKER,
            value_serializer=lambda v: json.dumps(v).encode("utf-8") if v else None,
            key_serializer=lambda v: json.dumps(v).encode("utf-8") if v else None,
        )
    elif client_type == "callable_object":
        producer = kafka.KafkaProducer(
            bootstrap_servers=BROKER,
            value_serializer=json_callable_serializer,
            key_serializer=json_callable_serializer,
        )
    elif client_type == "serializer_object":
        producer = kafka.KafkaProducer(
            bootstrap_servers=BROKER,
            value_serializer=json_serializer,
            key_serializer=json_serializer,
        )

    yield producer
    producer.close()


@pytest.fixture(scope="function")
def consumer(topic, producer, client_type, json_deserializer, json_callable_deserializer):
    if client_type == "no_serializer":
        consumer = kafka.KafkaConsumer(
            topic,
            bootstrap_servers=BROKER,
            auto_offset_reset="earliest",
            consumer_timeout_ms=100,
            heartbeat_interval_ms=1000,
            group_id="test",
        )
    elif client_type == "serializer_function":
        consumer = kafka.KafkaConsumer(
            topic,
            bootstrap_servers=BROKER,
            key_deserializer=lambda v: json.loads(v.decode("utf-8")) if v else None,
            value_deserializer=lambda v: json.loads(v.decode("utf-8")) if v else None,
            auto_offset_reset="earliest",
            consumer_timeout_ms=100,
            heartbeat_interval_ms=1000,
            group_id="test",
        )
    elif client_type == "callable_object":
        consumer = kafka.KafkaConsumer(
            topic,
            bootstrap_servers=BROKER,
            key_deserializer=json_callable_deserializer,
            value_deserializer=json_callable_deserializer,
            auto_offset_reset="earliest",
            consumer_timeout_ms=100,
            heartbeat_interval_ms=1000,
            group_id="test",
        )
    elif client_type == "serializer_object":
        consumer = kafka.KafkaConsumer(
            topic,
            bootstrap_servers=BROKER,
            key_deserializer=json_deserializer,
            value_deserializer=json_deserializer,
            auto_offset_reset="earliest",
            consumer_timeout_ms=100,
            heartbeat_interval_ms=1000,
            group_id="test",
        )

    yield consumer
    consumer.close()


@pytest.fixture(scope="session")
def serialize(client_type):
    if client_type == "no_serializer":
        return lambda v: json.dumps(v).encode("utf-8")
    else:
        return lambda v: v


@pytest.fixture(scope="session")
def deserialize(client_type):
    if client_type == "no_serializer":
        return lambda v: json.loads(v.decode("utf-8"))
    else:
        return lambda v: v


@pytest.fixture(scope="session")
def json_serializer():
    class JSONSerializer(kafka.serializer.Serializer):
        def serialize(self, topic, obj):
            return json.dumps(obj).encode("utf-8") if obj is not None else None

    return JSONSerializer()


@pytest.fixture(scope="session")
def json_deserializer():
    class JSONDeserializer(kafka.serializer.Deserializer):
        def deserialize(self, topic, bytes_):
            return json.loads(bytes_.decode("utf-8")) if bytes_ is not None else None

    return JSONDeserializer()


@pytest.fixture(scope="session")
def json_callable_serializer():
    class JSONCallableSerializer(object):
        def __call__(self, obj):
            return json.dumps(obj).encode("utf-8") if obj is not None else None

    return JSONCallableSerializer()


@pytest.fixture(scope="session")
def json_callable_deserializer():
    class JSONCallableDeserializer(object):
        def __call__(self, obj):
            return json.loads(obj.decode("utf-8")) if obj is not None else None

    return JSONCallableDeserializer()


@pytest.fixture(scope="function")
def topic():
    from kafka.admin.client import KafkaAdminClient
    from kafka.admin.new_topic import NewTopic

    topic = "test-topic-%s" % str(uuid.uuid4())

    admin = KafkaAdminClient(bootstrap_servers=BROKER)
    new_topics = [NewTopic(topic, num_partitions=1, replication_factor=1)]
    admin.create_topics(new_topics)

    yield topic

    admin.delete_topics([topic])


@pytest.fixture()
def send_producer_message(topic, producer, serialize):
    def _test():
        producer.send(topic, key=serialize("bar"), value=serialize({"foo": 1}))
        producer.flush()

    return _test


@pytest.fixture()
def get_consumer_record(topic, send_producer_message, consumer, deserialize):
    def _test():
        send_producer_message()

        record_count = 0

        timeout = 10
        attempts = 0
        record = None
        while not record and attempts < timeout:
            for record in consumer:
                assert deserialize(record.value) == {"foo": 1}
                record_count += 1
            attempts += 1

        assert record_count == 1, "Incorrect count of records consumed: %d. Expected 1." % record_count

    return _test


@transient_function_wrapper(kafka.producer.kafka, "KafkaProducer.send.__wrapped__")
# Place transient wrapper underneath instrumentation
def cache_kafka_producer_headers(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    ret = wrapped(*args, **kwargs)
    headers = kwargs.get("headers", [])
    headers = dict(headers)
    transaction._test_request_headers = headers
    return ret


@transient_function_wrapper(kafka.consumer.group, "KafkaConsumer.__next__")
# Place transient wrapper underneath instrumentation
def cache_kafka_consumer_headers(wrapped, instance, args, kwargs):
    record = wrapped(*args, **kwargs)
    transaction = current_transaction()

    if transaction is None:
        return record

    headers = record.headers
    headers = dict(headers)
    transaction._test_request_headers = headers
    return record


@pytest.fixture(autouse=True)
def assert_no_active_transaction():
    # Run before test
    assert not current_transaction(active_only=False), "Transaction exists before test run."

    yield  # Run test

    # Run after test
    assert not current_transaction(active_only=False), "Transaction was not properly exited."



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest
from testing_support.fixtures import (
    reset_core_stats_engine,
    validate_error_event_attributes_outside_transaction,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.object_names import callable_name
from newrelic.packages import six


def test_serialization_metrics(skip_if_not_serializing, topic, send_producer_message):
    txn_name = "test_serialization:test_serialization_metrics.<locals>.test" if six.PY3 else "test_serialization:test"

    _metrics = [
        ("MessageBroker/Kafka/Topic/Named/%s/Serialization/Value" % topic, 1),
        ("MessageBroker/Kafka/Topic/Named/%s/Serialization/Key" % topic, 1),
    ]

    @validate_transaction_metrics(
        txn_name,
        scoped_metrics=_metrics,
        rollup_metrics=_metrics,
        background_task=True,
    )
    @background_task()
    def test():
        send_producer_message()

    test()


@pytest.mark.parametrize(
    "key,value",
    (
        (object(), "A"),
        ("A", object()),
    ),
)
def test_serialization_errors(skip_if_not_serializing, topic, producer, key, value):
    error_cls = TypeError

    @validate_transaction_errors([callable_name(error_cls)])
    @background_task()
    def test():
        with pytest.raises(error_cls):
            producer.send(topic=topic, key=key, value=value)

    test()


@pytest.mark.parametrize(
    "key,value",
    (
        (b"%", b"{}"),
        (b"{}", b"%"),
    ),
)
def test_deserialization_errors(skip_if_not_serializing, monkeypatch, topic, producer, consumer, key, value):
    error_cls = json.decoder.JSONDecodeError if six.PY3 else ValueError

    # Remove serializers to cause intentional issues
    monkeypatch.setitem(producer.config, "value_serializer", None)
    monkeypatch.setitem(producer.config, "key_serializer", None)

    producer.send(topic=topic, key=key, value=value)
    producer.flush()

    @reset_core_stats_engine()
    @validate_error_event_attributes_outside_transaction(
        num_errors=1, exact_attrs={"intrinsic": {"error.class": callable_name(error_cls)}, "agent": {}, "user": {}}
    )
    def test():
        with pytest.raises(error_cls):
            timeout = 10
            attempts = 0
            record = None
            while not record and attempts < timeout:
                for record in consumer:
                    pass
                attempts += 1

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from testing_support.fixtures import (
    core_application_stats_engine_error,
    error_is_saved,
    override_application_settings,
    reset_core_stats_engine,
    validate_transaction_error_event_count,
    validate_transaction_error_trace_count,
)
from testing_support.validators.validate_application_error_event_count import (
    validate_application_error_event_count,
)
from testing_support.validators.validate_application_error_trace_count import (
    validate_application_error_trace_count,
)
from testing_support.validators.validate_application_errors import (
    validate_application_errors,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)

from newrelic.api.application import application_instance as application
from newrelic.api.application import application_settings
from newrelic.api.background_task import background_task
from newrelic.api.settings import STRIP_EXCEPTION_MESSAGE
from newrelic.api.time_trace import notice_error
from newrelic.common.object_names import callable_name

_runtime_error_name = callable_name(RuntimeError)
_type_error_name = callable_name(TypeError)

# =============== Test errors during a transaction ===============

_test_notice_error_sys_exc_info = [(_runtime_error_name, "one")]


@validate_transaction_errors(errors=_test_notice_error_sys_exc_info)
@background_task()
def test_notice_error_sys_exc_info():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error(sys.exc_info())


_test_notice_error_no_exc_info = [(_runtime_error_name, "one")]


@validate_transaction_errors(errors=_test_notice_error_no_exc_info)
@background_task()
def test_notice_error_no_exc_info():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error()


_test_notice_error_custom_params = [(_runtime_error_name, "one")]


@validate_transaction_errors(errors=_test_notice_error_custom_params, required_params=[("key", "value")])
@background_task()
def test_notice_error_custom_params():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error(sys.exc_info(), attributes={"key": "value"})


_test_notice_error_multiple_different_type = [(_runtime_error_name, "one"), (_type_error_name, "two")]


@validate_transaction_errors(errors=_test_notice_error_multiple_different_type)
@background_task()
def test_notice_error_multiple_different_type():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error()

    try:
        raise TypeError("two")
    except TypeError:
        notice_error()


_test_notice_error_multiple_same_type = [(_runtime_error_name, "one"), (_runtime_error_name, "two")]


@validate_transaction_errors(errors=_test_notice_error_multiple_same_type)
@background_task()
def test_notice_error_multiple_same_type():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error()

    try:
        raise RuntimeError("two")
    except RuntimeError:
        notice_error()


# =============== Test errors outside a transaction ===============

_test_application_exception = [(_runtime_error_name, "one")]


@reset_core_stats_engine()
@validate_application_errors(errors=_test_application_exception)
def test_application_exception():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        application_instance = application()
        notice_error(application=application_instance)


_test_application_exception_sys_exc_info = [(_runtime_error_name, "one")]


@reset_core_stats_engine()
@validate_application_errors(errors=_test_application_exception_sys_exc_info)
def test_application_exception_sys_exec_info():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        application_instance = application()
        notice_error(sys.exc_info(), application=application_instance)


_test_application_exception_custom_params = [(_runtime_error_name, "one")]


@reset_core_stats_engine()
@validate_application_errors(errors=_test_application_exception_custom_params, required_params=[("key", "value")])
def test_application_exception_custom_params():
    try:
        raise RuntimeError("one")
    except RuntimeError:
        application_instance = application()
        notice_error(attributes={"key": "value"}, application=application_instance)


_test_application_exception_multiple = [(_runtime_error_name, "one"), (_runtime_error_name, "one")]


@reset_core_stats_engine()
@validate_application_errors(errors=_test_application_exception_multiple)
@background_task()
def test_application_exception_multiple():
    """Exceptions submitted straight to the stats engine doesn't check for
    duplicates
    """
    application_instance = application()
    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error(application=application_instance)

    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error(application=application_instance)


# =============== Test exception message stripping/allowlisting ===============

_test_notice_error_strip_message_disabled = [(_runtime_error_name, "one")]

_strip_message_disabled_settings = {
    "strip_exception_messages.enabled": False,
}


@validate_transaction_errors(errors=_test_notice_error_strip_message_disabled)
@override_application_settings(_strip_message_disabled_settings)
@background_task()
def test_notice_error_strip_message_disabled():
    settings = application_settings()
    assert not settings.strip_exception_messages.enabled

    try:
        raise RuntimeError("one")
    except RuntimeError:
        notice_error()


class ErrorOne(Exception):
    message = "error one message"


_error_one_name = callable_name(ErrorOne)


@override_application_settings(_strip_message_disabled_settings)
def test_notice_error_strip_message_disabled_outside_transaction():
    settings = application_settings()
    assert not settings.strip_exception_messages.enabled

    try:
        assert not error_is_saved(ErrorOne)
        raise ErrorOne(ErrorOne.message)
    except ErrorOne:
        application_instance = application()
        application_instance.notice_error()

    my_error = core_application_stats_engine_error(_error_one_name)
    assert my_error.message == ErrorOne.message


_test_notice_error_strip_message_enabled = [(_runtime_error_name, STRIP_EXCEPTION_MESSAGE)]

_strip_message_enabled_settings = {
    "strip_exception_messages.enabled": True,
}


@validate_transaction_errors(errors=_test_notice_error_strip_message_enabled)
@override_application_settings(_strip_message_enabled_settings)
@background_task()
def test_notice_error_strip_message_enabled():
    settings = application_settings()
    assert settings.strip_exception_messages.enabled

    try:
        raise RuntimeError("message not displayed")
    except RuntimeError:
        notice_error()


class ErrorTwo(Exception):
    message = "error two message"


_error_two_name = callable_name(ErrorTwo)


@override_application_settings(_strip_message_enabled_settings)
def test_notice_error_strip_message_enabled_outside_transaction():
    settings = application_settings()
    assert settings.strip_exception_messages.enabled

    try:
        assert not error_is_saved(ErrorTwo)
        raise ErrorTwo(ErrorTwo.message)
    except ErrorTwo:
        application_instance = application()
        application_instance.notice_error()

    my_error = core_application_stats_engine_error(_error_two_name)
    assert my_error.message == STRIP_EXCEPTION_MESSAGE


_test_notice_error_strip_message_in_allowlist = [(_runtime_error_name, "original error message")]

_strip_message_in_allowlist_settings = {
    "strip_exception_messages.enabled": True,
    "strip_exception_messages.allowlist": [_runtime_error_name],
}


@validate_transaction_errors(errors=_test_notice_error_strip_message_in_allowlist)
@override_application_settings(_strip_message_in_allowlist_settings)
@background_task()
def test_notice_error_strip_message_in_allowlist():
    settings = application_settings()
    assert settings.strip_exception_messages.enabled
    assert _runtime_error_name in settings.strip_exception_messages.allowlist

    try:
        raise RuntimeError("original error message")
    except RuntimeError:
        notice_error()


class ErrorThree(Exception):
    message = "error three message"


_error_three_name = callable_name(ErrorThree)

_strip_message_in_allowlist_settings_outside_transaction = {
    "strip_exception_messages.enabled": True,
    "strip_exception_messages.allowlist": [_error_three_name],
}


@override_application_settings(_strip_message_in_allowlist_settings_outside_transaction)
def test_notice_error_strip_message_in_allowlist_outside_transaction():
    settings = application_settings()
    assert settings.strip_exception_messages.enabled
    assert _error_three_name in settings.strip_exception_messages.allowlist

    try:
        assert not error_is_saved(ErrorThree)
        raise ErrorThree(ErrorThree.message)
    except ErrorThree:
        application_instance = application()
        application_instance.notice_error()

    my_error = core_application_stats_engine_error(_error_three_name)
    assert my_error.message == ErrorThree.message


_test_notice_error_strip_message_not_in_allowlist = [(_runtime_error_name, STRIP_EXCEPTION_MESSAGE)]

_strip_message_not_in_allowlist_settings = {
    "strip_exception_messages.enabled": True,
    "strip_exception_messages.allowlist": ["FooError", "BarError"],
}


@validate_transaction_errors(errors=_test_notice_error_strip_message_not_in_allowlist)
@override_application_settings(_strip_message_not_in_allowlist_settings)
@background_task()
def test_notice_error_strip_message_not_in_allowlist():
    settings = application_settings()
    assert settings.strip_exception_messages.enabled
    assert _runtime_error_name not in settings.strip_exception_messages.allowlist

    try:
        raise RuntimeError("message not displayed")
    except RuntimeError:
        notice_error()


class ErrorFour(Exception):
    message = "error four message"


_error_four_name = callable_name(ErrorFour)

_strip_message_not_in_allowlist_settings_outside_transaction = {
    "strip_exception_messages.enabled": True,
    "strip_exception_messages.allowlist": ["ValueError", "BarError"],
}


@override_application_settings(_strip_message_not_in_allowlist_settings_outside_transaction)
def test_notice_error_strip_message_not_in_allowlist_outside_transaction():
    settings = application_settings()
    assert settings.strip_exception_messages.enabled
    assert _error_four_name not in settings.strip_exception_messages.allowlist

    try:
        assert not error_is_saved(ErrorFour)
        raise ErrorFour(ErrorFour.message)
    except ErrorFour:
        application_instance = application()
        application_instance.notice_error()

    my_error = core_application_stats_engine_error(_error_four_name)
    assert my_error.message == STRIP_EXCEPTION_MESSAGE


# =============== Test exception limits ===============


def _raise_errors(num_errors, application=None):
    for i in range(num_errors):
        try:
            raise RuntimeError("error" + str(i))
        except RuntimeError:
            notice_error(application=application)


_errors_per_transaction_limit = 5
_num_errors_transaction = 6
_errors_per_harvest_limit = 20
_num_errors_app = 26
_error_event_limit = 25


@override_application_settings({"agent_limits.errors_per_transaction": _errors_per_transaction_limit})
@validate_transaction_error_trace_count(_errors_per_transaction_limit)
@background_task()
def test_transaction_error_trace_limit():
    _raise_errors(_num_errors_transaction)


@override_application_settings({"agent_limits.errors_per_harvest": _errors_per_harvest_limit})
@reset_core_stats_engine()
@validate_application_error_trace_count(_errors_per_harvest_limit)
def test_application_error_trace_limit():
    _raise_errors(_num_errors_app, application())


# The limit for errors on transactions is shared for traces and errors


@override_application_settings(
    {
        "agent_limits.errors_per_transaction": _errors_per_transaction_limit,
        "error_collector.max_event_samples_stored": _error_event_limit,
    }
)
@validate_transaction_error_event_count(_errors_per_transaction_limit)
@background_task()
def test_transaction_error_event_limit():
    _raise_errors(_num_errors_transaction)


# The harvest limit for error traces doesn't affect events


@override_application_settings(
    {
        "agent_limits.errors_per_harvest": _errors_per_harvest_limit,
        "event_harvest_config.harvest_limits.error_event_data": _error_event_limit,
    }
)
@reset_core_stats_engine()
@validate_application_error_event_count(_error_event_limit)
def test_application_error_event_limit():
    _raise_errors(_num_errors_app, application())


# =============== Test params is not a dict ===============


@reset_core_stats_engine()
@validate_transaction_error_trace_count(num_errors=1)
@background_task()
def test_transaction_notice_error_params_not_a_dict():
    try:
        raise RuntimeError()
    except RuntimeError:
        notice_error(sys.exc_info(), attributes=[1, 2, 3])


@reset_core_stats_engine()
@validate_application_error_trace_count(num_errors=1)
def test_application_notice_error_params_not_a_dict():
    try:
        raise RuntimeError()
    except RuntimeError:
        notice_error(sys.exc_info(), attributes=[1, 2, 3], application=application())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import aredis

from newrelic.api.transaction import current_transaction
from newrelic.api.background_task import background_task

from testing_support.db_settings import redis_settings
from testing_support.fixture.event_loop import event_loop as loop
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.util import instance_hostname


DB_SETTINGS = redis_settings()[0]
DATABASE_NUMBER = 0


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
    'datastore_tracer.database_name_reporting.enabled': True,
    'distributed_tracing.enabled': True,
    'span_events.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
    'datastore_tracer.database_name_reporting.enabled': False,
    'distributed_tracing.enabled': True,
    'span_events.enabled': True,
}


async def _exercise_db():
    client = aredis.StrictRedis(host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'], db=DATABASE_NUMBER)

    await client.set('key', 'value')
    await client.get('key')

    await client.execute_command('CLIENT', 'LIST', parse='LIST')


# Tests

@pytest.mark.parametrize('db_instance_enabled', (True, False))
@pytest.mark.parametrize('instance_enabled', (True, False))
def test_span_events(instance_enabled, db_instance_enabled, loop):
    guid = 'dbb533c53b749e0b'
    priority = 0.5

    common = {
        'type': 'Span',
        'transactionId': guid,
        'priority': priority,
        'sampled': True,
        'category': 'datastore',
        'component': 'Redis',
        'span.kind': 'client',
    }
    exact_agents = {}

    if instance_enabled:
        settings = _enable_instance_settings.copy()
        hostname = instance_hostname(DB_SETTINGS['host'])
        exact_agents.update({
            'peer.address': '%s:%s' % (hostname, DB_SETTINGS['port']),
            'peer.hostname': hostname,
        })
    else:
        settings = _disable_instance_settings.copy()
        exact_agents.update({
            'peer.address': 'Unknown:Unknown',
            'peer.hostname': 'Unknown',
        })

    if db_instance_enabled and instance_enabled:
        exact_agents.update({
            'db.instance': str(DATABASE_NUMBER),
        })
        unexpected_agents = ()
    else:
        settings['attributes.exclude'] = ['db.instance']
        unexpected_agents = ('db.instance',)

    query_1 = common.copy()
    query_1['name'] = 'Datastore/operation/Redis/set'

    query_2 = common.copy()
    query_2['name'] = 'Datastore/operation/Redis/get'

    query_3 = common.copy()
    query_3['name'] = 'Datastore/operation/Redis/client_list'

    @validate_span_events(
            count=1,
            exact_intrinsics=query_1,
            unexpected_intrinsics=('db.instance'),
            exact_agents=exact_agents,
            unexpected_agents=unexpected_agents)
    @validate_span_events(
            count=1,
            exact_intrinsics=query_2,
            unexpected_intrinsics=('db.instance'),
            exact_agents=exact_agents,
            unexpected_agents=unexpected_agents)
    @validate_span_events(
            count=1,
            exact_intrinsics=query_3,
            unexpected_intrinsics=('db.instance'),
            exact_agents=exact_agents,
            unexpected_agents=unexpected_agents)
    @override_application_settings(settings)
    @background_task(name='span_events')
    def _test():
        txn = current_transaction()
        txn.guid = guid
        txn._priority = priority
        txn._sampled = True
        loop.run_until_complete(_exercise_db())

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import socket
import sys

import pytest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.mock_external_http_server import (
    MockExternalHTTPHResponseHeadersServer,
    MockExternalHTTPServer,
)
from testing_support.validators.validate_distributed_tracing_header import (
    validate_distributed_tracing_header,
)
from testing_support.validators.validate_outbound_headers import (
    validate_outbound_headers,
)

from newrelic.api.background_task import background_task
from newrelic.api.function_trace import FunctionTrace
from newrelic.api.transaction import current_transaction

ENCODING_KEY = "1234567890123456789012345678901234567890"

is_pypy = hasattr(sys, "pypy_version_info")


@pytest.fixture(scope="module")
def external():
    external = MockExternalHTTPHResponseHeadersServer()
    with external:
        yield external


@background_task(name="make_request")
def make_request(port, req_type, client_cls, count=1, raise_error=True, as_kwargs=True, **kwargs):
    import tornado.concurrent
    import tornado.gen
    import tornado.httpclient
    import tornado.ioloop

    class CustomAsyncHTTPClient(tornado.httpclient.AsyncHTTPClient):
        def fetch_impl(self, request, callback):
            out = []
            for k, v in request.headers.items():
                out.append("%s: %s" % (k, v))
            body = "\n".join(out).encode("utf-8")
            response = tornado.httpclient.HTTPResponse(request=request, code=200, buffer=io.BytesIO(body))
            callback(response)

    cls = tornado.httpclient.AsyncHTTPClient
    if client_cls == "AsyncHTTPClient":
        cls.configure(None)
    elif client_cls == "CurlAsyncHTTPClient":
        if is_pypy:
            pytest.skip("Pypy not supported by Pycurl")
        import tornado.curl_httpclient

        cls.configure(tornado.curl_httpclient.CurlAsyncHTTPClient)
    elif client_cls == "HTTPClient":
        cls = tornado.httpclient.HTTPClient
    elif client_cls == "CustomAsyncHTTPClient":
        cls.configure(CustomAsyncHTTPClient)
    else:
        raise ValueError("Received unknown client type: %s" % client_cls)

    client = cls(force_instance=True)
    callback = None

    uri = "http://localhost:%s" % port
    if req_type == "class":
        req = tornado.httpclient.HTTPRequest(uri, **kwargs)
        kwargs = {}
    elif req_type == "uri":
        req = uri
    else:
        raise ValueError("Received unknown request type: %s" % req_type)

    @tornado.gen.coroutine
    def _make_request():

        if as_kwargs:
            futures = [client.fetch(req, raise_error=raise_error, **kwargs) for _ in range(count)]
        elif tornado.version_info < (6, 0):
            futures = [client.fetch(req, callback, raise_error, **kwargs) for _ in range(count)]
        else:
            futures = [client.fetch(req, raise_error, **kwargs) for _ in range(count)]

        assert all(isinstance(f, tornado.concurrent.Future) for f in futures)

        if count > 1:
            responses = yield tornado.gen.multi_future(futures)
            response = responses[0]
        else:
            response = yield futures[0]

        raise tornado.gen.Return(response)

    if client_cls == "HTTPClient":
        for _ in range(count):
            response = client.fetch(req, raise_error=raise_error, **kwargs)
        return response
    else:
        response = tornado.ioloop.IOLoop.current().run_sync(_make_request)
        return response


@pytest.mark.parametrize(
    "client_class,as_kwargs",
    [
        ("AsyncHTTPClient", True),
        ("AsyncHTTPClient", False),
        ("CurlAsyncHTTPClient", True),
        ("CurlAsyncHTTPClient", False),
        ("CustomAsyncHTTPClient", True),
        ("CustomAsyncHTTPClient", False),
        ("HTTPClient", True),
    ],
)
@pytest.mark.parametrize(
    "cat_enabled,user_header,span_events,distributed_tracing",
    [
        (True, None, False, False),
        (True, "X-NewRelic-ID", False, False),
        (True, "X-NewRelic-Transaction", False, False),
        (False, None, True, True),
        (False, None, False, True),
    ],
)
# @pytest.mark.parametrize('cat_enabled,user_header', [
#    (True, None),
#    (True, 'X-NewRelic-ID'),
#    (True, 'X-NewRelic-Transaction'),
#    (False, None),
# ])
@pytest.mark.parametrize("request_type", ["uri", "class"])
@pytest.mark.parametrize("num_requests", [1, 2])
def test_httpclient(
    cat_enabled,
    request_type,
    client_class,
    user_header,
    num_requests,
    distributed_tracing,
    span_events,
    external,
    as_kwargs,
):

    port = external.port

    expected_metrics = [("External/localhost:%s/tornado/GET" % port, num_requests)]

    @override_application_settings(
        {
            "distributed_tracing.enabled": distributed_tracing,
            "span_events.enabled": span_events,
            "cross_application_tracer.enabled": not distributed_tracing,
        }
    )
    @validate_transaction_metrics(
        "test_externals:test_httpclient",
        background_task=True,
        rollup_metrics=expected_metrics,
        scoped_metrics=expected_metrics,
    )
    @background_task(name="test_externals:test_httpclient")
    def _test():
        headers = {}
        if user_header:
            headers = {user_header: "USER"}

        response = make_request(
            port, request_type, client_class, headers=headers, count=num_requests, as_kwargs=as_kwargs
        )
        assert response.code == 200

        body = response.body
        if hasattr(body, "decode"):
            body = body.decode("utf-8")

        headers = {}
        for header_line in body.split("\n"):
            if ":" not in header_line:
                continue
            header_key, header_val = header_line.split(":", 1)
            header_key = header_key.strip()
            header_val = header_val.strip()
            headers[header_key] = header_val

        # User headers override all inserted NR headers
        if user_header:
            assert headers[user_header] == "USER"
        elif cat_enabled:
            t = current_transaction()
            assert t
            t._test_request_headers = headers

            if distributed_tracing:
                validate_distributed_tracing_header(header="Newrelic")
            else:
                validate_outbound_headers()
        else:
            # new relic shouldn't add anything to the outgoing
            assert "x-newrelic" not in body, body

        assert "X-NewRelic-App-Data" not in headers

    _test()


CAT_RESPONSE_CODE = None


def cat_response_handler(self):
    global CAT_RESPONSE_CODE
    # payload
    # (
    #     u'1#1', u'WebTransaction/Function/app:beep',
    #     0, 1.23, -1,
    #     'dd4a810b7cb7f937',
    #     False,
    # )
    cat_response_header = (
        "X-NewRelic-App-Data",
        "ahACFwQUGxpuVVNmQVVbRVZbTVleXBxyQFhUTFBfXx1SREUMVV1cQBMeAxgEGAULFR0AHhFQUQJWAAgAUwVQVgJQDgsOEh1UUlhGU2o=",
    )
    self.send_response(CAT_RESPONSE_CODE)
    self.send_header(*cat_response_header)
    self.end_headers()
    self.wfile.write(b"Example Data")


@pytest.fixture(scope="module")
def cat_response_server():
    external = MockExternalHTTPServer(handler=cat_response_handler)
    with external:
        yield external


@pytest.mark.parametrize("client_class", ["AsyncHTTPClient", "CurlAsyncHTTPClient", "HTTPClient"])
@pytest.mark.parametrize("cat_enabled", [True, False])
@pytest.mark.parametrize("request_type", ["uri", "class"])
@pytest.mark.parametrize(
    "response_code,raise_error",
    [
        (500, True),
        (500, False),
        (200, False),
    ],
)
def test_client_cat_response_processing(
    cat_enabled, request_type, client_class, raise_error, response_code, cat_response_server
):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = response_code

    _custom_settings = {
        "cross_process_id": "1#1",
        "encoding_key": ENCODING_KEY,
        "trusted_account_ids": [1],
        "cross_application_tracer.enabled": cat_enabled,
        "distributed_tracing.enabled": not cat_enabled,
        "transaction_tracer.transaction_threshold": 0.0,
    }

    port = cat_response_server.port
    expected_metrics = [
        ("ExternalTransaction/localhost:%s/1#1/WebTransaction/Function/app:beep" % port, 1 if cat_enabled else None),
    ]

    @validate_transaction_metrics(
        "make_request", background_task=True, rollup_metrics=expected_metrics, scoped_metrics=expected_metrics
    )
    @override_application_settings(_custom_settings)
    def _test():
        import tornado
        import tornado.httpclient

        try:
            response = make_request(port, request_type, client_class, raise_error=raise_error)
        except tornado.httpclient.HTTPError as e:
            assert raise_error
            response = e.response
        else:
            assert not raise_error

        assert response.code == response_code

    _test()


@pytest.mark.parametrize("client_class", ["AsyncHTTPClient", "CurlAsyncHTTPClient", "HTTPClient"])
@validate_transaction_metrics("make_request", background_task=True)
def test_httpclient_invalid_method(client_class, external):
    with pytest.raises(KeyError):
        make_request(external.port, "uri", client_class, method="COOKIES")


@pytest.mark.parametrize("client_class", ["AsyncHTTPClient", "CurlAsyncHTTPClient", "HTTPClient"])
@validate_transaction_metrics("make_request", background_task=True)
def test_httpclient_invalid_kwarg(client_class, external):
    with pytest.raises(TypeError):
        make_request(external.port, "uri", client_class, boop="1234")


def test_httpclient_fetch_crashes(external):
    @validate_transaction_metrics(
        "test_httpclient_fetch_crashes",
        background_task=True,
        rollup_metrics=[("External/localhost:%d/tornado/GET" % external.port, 1)],
        scoped_metrics=[("External/localhost:%d/tornado/GET" % external.port, 1)],
    )
    @background_task(name="test_httpclient_fetch_crashes")
    def _test():
        import tornado.httpclient

        class CrashClient(tornado.httpclient.AsyncHTTPClient):
            def fetch_impl(self, *args, **kwargs):
                raise ValueError("BOOM")

        client = CrashClient(force_instance=True)

        port = external.port
        with pytest.raises(ValueError):
            tornado.ioloop.IOLoop.current().run_sync(lambda: client.fetch("http://localhost:%s" % port))

    _test()


def test_httpclient_fetch_inside_terminal_node(external):
    @validate_transaction_metrics(
        "test_httpclient_fetch_inside_terminal_node",
        background_task=True,
        rollup_metrics=[("External/localhost:%d/tornado/GET" % external.port, None)],
        scoped_metrics=[("External/localhost:%d/tornado/GET" % external.port, None)],
    )
    @background_task(name="test_httpclient_fetch_inside_terminal_node")
    def _test():
        # Test that our instrumentation correctly handles the case when the parent
        # is a terminal node
        import tornado.gen
        import tornado.httpclient
        import tornado.ioloop

        client = tornado.httpclient.AsyncHTTPClient(force_instance=True)

        # This is protecting against a "pop_current" when the external trace never
        # actually gets pushed
        port = external.port

        @tornado.gen.coroutine
        def _make_request():
            with FunctionTrace(name="parent", terminal=True):
                response = yield client.fetch("http://localhost:%s" % port)
            return response

        response = tornado.ioloop.IOLoop.current().run_sync(_make_request)
        assert response.code == 200

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.core.rules_engine import RulesEngine, NormalizationRule

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.normpath(os.path.join(
        CURRENT_DIR, 'fixtures', 'rules.json'))

def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)

def _prepare_rules(test_rules):
    # ensure all keys are present, if not present set to an empty string
    for rule in test_rules:
        for key in NormalizationRule._fields:
            rule[key] = rule.get(key, '')
    return test_rules

def _make_case_insensitive(rules):
    # lowercase each rule
    for rule in rules:
        rule['match_expression'] = rule['match_expression'].lower()
        if rule.get('replacement'):
            rule['replacement'] = rule['replacement'].lower()
    return rules

@pytest.mark.parametrize('test_group', _load_tests())
def test_rules_engine(test_group):

    # FIXME: The test fixture assumes that matching is case insensitive when it
    # is not. To avoid errors, just lowercase all rules, inputs, and expected
    # values.
    insense_rules = _make_case_insensitive(test_group['rules'])
    test_rules = _prepare_rules(insense_rules)
    rules_engine = RulesEngine(test_rules)

    for test in test_group['tests']:

        # lowercase each value
        input_str = test['input'].lower()
        expected = (test['expected'] or '').lower()

        result, ignored = rules_engine.normalize(input_str)

        # When a transaction is to be ignored, the test fixture expects that
        # "expected" is None.
        if ignored:
            assert expected == ''
        else:
            assert result == expected



# -*- coding: utf-8 -*-
# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import pytest

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.time_trace import notice_error

from testing_support.fixtures import (validate_transaction_exception_message,
        set_default_encoding, validate_application_exception_message,
        reset_core_stats_engine)

UNICODE_MESSAGE = u'I💜🐍'
UNICODE_ENGLISH = u'I love python'
BYTES_ENGLISH = b'I love python'
BYTES_UTF8_ENCODED = b'I\xf0\x9f\x92\x9c\xf0\x9f\x90\x8d'
INCORRECTLY_DECODED_BYTES_PY2 = u'I\u00f0\u009f\u0092\u009c\u00f0\u009f\u0090\u008d'
INCORRECTLY_DECODED_BYTES_PY3 = u"b'I\\xf0\\x9f\\x92\\x9c\\xf0\\x9f\\x90\\x8d'"
# =================== Exception messages during transaction ====================

# ---------------- Python 2

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('ascii')
@validate_transaction_exception_message(UNICODE_MESSAGE)
@background_task()
def test_py2_transaction_exception_message_unicode():
    """Assert unicode message when using non-ascii characters is preserved,
    with sys default encoding"""
    try:
        raise ValueError(UNICODE_MESSAGE)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('ascii')
@validate_transaction_exception_message(UNICODE_ENGLISH)
@background_task()
def test_py2_transaction_exception_message_unicode_english():
    """Assert unicode message when using ascii compatible characters preserved,
    with sys default encoding"""
    try:
        raise ValueError(UNICODE_ENGLISH)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('ascii')
@validate_transaction_exception_message(UNICODE_ENGLISH)
@background_task()
def test_py2_transaction_exception_message_bytes_english():
    """Assert byte string of ascii characters decodes sensibly"""
    try:
        raise ValueError(BYTES_ENGLISH)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('ascii')
@validate_transaction_exception_message(INCORRECTLY_DECODED_BYTES_PY2)
@background_task()
def test_py2_transaction_exception_message_bytes_non_english():
    """Assert known situation where (explicitly) utf-8 encoded byte string gets
    mangled when default sys encoding is ascii. THIS TEST ASSERTS THAT THE
    MESSAGE IS WRONG. We do not expect it to work now, or in the future.
    """
    try:
        raise ValueError(BYTES_UTF8_ENCODED)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('ascii')
@validate_transaction_exception_message(INCORRECTLY_DECODED_BYTES_PY2)
@background_task()
def test_py2_transaction_exception_message_bytes_implicit_encoding_non_english():
    """Assert known situation where (implicitly) utf-8 encoded byte string gets
    mangled when default sys encoding is ascii. THIS TEST ASSERTS THAT THE
    MESSAGE IS WRONG. We do not expect it to work now, or in the future.
    """
    try:

        # Bytes literal with non-ascii compatible characters only allowed in
        # python 2

        raise ValueError('I💜🐍')
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('utf-8')
@validate_transaction_exception_message(UNICODE_MESSAGE)
@background_task()
def test_py2_transaction_exception_message_unicode_utf8_encoding():
    """Assert unicode error message is preserved with sys non-default utf-8
    encoding
    """
    try:
        raise ValueError(UNICODE_MESSAGE)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@set_default_encoding('utf-8')
@validate_transaction_exception_message(UNICODE_MESSAGE)
@background_task()
def test_py2_transaction_exception_message_bytes_utf8_encoding_non_english():
    """Assert utf-8 encoded byte produces correct exception message when sys
    encoding is also utf-8.
    """
    try:

        # Bytes literal with non-ascii compatible characters only allowed in
        # python 2

        raise ValueError('I💜🐍')
    except ValueError:
        notice_error()

# ---------------- Python 3

@pytest.mark.skipif(six.PY2, reason="Testing Python 3 string behavior")
@validate_transaction_exception_message(UNICODE_MESSAGE)
@background_task()
def test_py3_transaction_exception_message_bytes_non_english_unicode():
    """Assert (native) unicode exception message is preserved when when
    non-ascii compatible characters present"""
    try:
        raise ValueError(UNICODE_MESSAGE)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY2, reason="Testing Python 3 string behavior")
@validate_transaction_exception_message(UNICODE_ENGLISH)
@background_task()
def test_py3_transaction_exception_message_unicode_english():
    """Assert (native) unicode exception message is preserved, when characters
    are ascii-compatible"""
    try:
        raise ValueError(UNICODE_ENGLISH)
    except ValueError:
        notice_error()

@pytest.mark.skipif(six.PY2, reason="Testing Python 3 string behavior")
@validate_transaction_exception_message(INCORRECTLY_DECODED_BYTES_PY3)
@background_task()
def test_py3_transaction_exception_message_bytes_non_english():
    """An issue can occur if you cast from bytes to a string in
    python 3 (that is using str(), not using encode/decode methods).
    This is because all characters in bytes are literals, no implicit
    decoding happens, like it does in python 2. You just shouldn't use bytes
    for exception messages in Python 3. THIS TEST ASSERTS THAT THE
    MESSAGE IS WRONG. We do not expect it to work now, or in the future.
    """
    try:
        raise ValueError(BYTES_UTF8_ENCODED)
    except ValueError:
        notice_error()

# =================== Exception messages outside transaction ====================

# ---------------- Python 2

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('ascii')
@validate_application_exception_message(UNICODE_MESSAGE)
def test_py2_application_exception_message_unicode():
    """Assert unicode message when using non-ascii characters is preserved,
    with sys default encoding"""
    try:
        raise ValueError(UNICODE_MESSAGE)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('ascii')
@validate_application_exception_message(UNICODE_ENGLISH)
def test_py2_application_exception_message_unicode_english():
    """Assert unicode message when using ascii compatible characters preserved,
    with sys default encoding"""
    try:
        raise ValueError(UNICODE_ENGLISH)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('ascii')
@validate_application_exception_message(UNICODE_ENGLISH)
def test_py2_application_exception_message_bytes_english():
    """Assert byte string of ascii characters decodes sensibly"""
    try:
        raise ValueError(BYTES_ENGLISH)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('ascii')
@validate_application_exception_message(INCORRECTLY_DECODED_BYTES_PY2)
def test_py2_application_exception_message_bytes_non_english():
    """Assert known situation where (explicitly) utf-8 encoded byte string gets
    mangled when default sys encoding is ascii. THIS TEST ASSERTS THAT THE
    MESSAGE IS WRONG. We do not expect it to work now, or in the future.
    """
    try:
        raise ValueError(BYTES_UTF8_ENCODED)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('ascii')
@validate_application_exception_message(INCORRECTLY_DECODED_BYTES_PY2)
def test_py2_application_exception_message_bytes_implicit_encoding_non_english():
    """Assert known situation where (implicitly) utf-8 encoded byte string gets
    mangled when default sys encoding is ascii. THIS TEST ASSERTS THAT THE
    MESSAGE IS WRONG. We do not expect it to work now, or in the future.
    """
    try:

        # Bytes literal with non-ascii compatible characters only allowed in
        # python 2

        raise ValueError('I💜🐍')
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('utf-8')
@validate_application_exception_message(UNICODE_MESSAGE)
def test_py2_application_exception_message_unicode_utf8_encoding():
    """Assert unicode error message is preserved with sys non-default utf-8
    encoding
    """
    try:
        raise ValueError(UNICODE_MESSAGE)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY3, reason="Testing Python 2 string behavior")
@reset_core_stats_engine()
@set_default_encoding('utf-8')
@validate_application_exception_message(UNICODE_MESSAGE)
def test_py2_application_exception_message_bytes_utf8_encoding_non_english():
    """Assert utf-8 encoded byte produces correct exception message when sys
    encoding is also utf-8.
    """
    try:

        # Bytes literal with non-ascii compatible characters only allowed in
        # python 2

        raise ValueError('I💜🐍')
    except ValueError:
        app = application()
        notice_error(application=app)

# ---------------- Python 3

@pytest.mark.skipif(six.PY2, reason="Testing Python 3 string behavior")
@reset_core_stats_engine()
@validate_application_exception_message(UNICODE_MESSAGE)
def test_py3_application_exception_message_bytes_non_english_unicode():
    """Assert (native) unicode exception message is preserved when when
    non-ascii compatible characters present"""
    try:
        raise ValueError(UNICODE_MESSAGE)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY2, reason="Testing Python 3 string behavior")
@reset_core_stats_engine()
@validate_application_exception_message(UNICODE_ENGLISH)
def test_py3_application_exception_message_unicode_english():
    """Assert (native) unicode exception message is preserved, when characters
    are ascii-compatible"""
    try:
        raise ValueError(UNICODE_ENGLISH)
    except ValueError:
        app = application()
        notice_error(application=app)

@pytest.mark.skipif(six.PY2, reason="Testing Python 3 string behavior")
@reset_core_stats_engine()
@validate_application_exception_message(INCORRECTLY_DECODED_BYTES_PY3)
def test_py3_application_exception_message_bytes_non_english():
    """It really makes a mess of things when you cast from bytes to a
    string in python 3 (that is using str(), not using encode/decode methods).
    This is because all characters in bytes are literals, no implicit
    decoding happens, like it does in python 2. You just shouldn't use bytes
    for exception messages in Python 3. THIS TEST ASSERTS THAT THE
    MESSAGE IS WRONG. We do not expect it to work now, or in the future.
    """
    try:
        raise ValueError(BYTES_UTF8_ENCODED)
    except ValueError:
        app = application()
        notice_error(application=app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_count import validate_transaction_count

from newrelic.api.background_task import background_task
from newrelic.common.object_names import callable_name


@pytest.fixture(scope="session")
def is_graphql_2():
    from graphql import __version__ as version

    major_version = int(version.split(".")[0])
    return major_version == 2


@pytest.fixture(scope="session")
def graphql_run():
    """Wrapper function to simulate framework_graphql test behavior."""

    def execute(schema, *args, **kwargs):
        return schema.execute_sync(*args, **kwargs)

    return execute


def to_graphql_source(query):
    def delay_import():
        try:
            from graphql import Source
        except ImportError:
            # Fallback if Source is not implemented
            return query

        from graphql import __version__ as version

        # For graphql2, Source objects aren't acceptable input
        major_version = int(version.split(".")[0])
        if major_version == 2:
            return query

        return Source(query)

    return delay_import


def example_middleware(next, root, info, **args):  #pylint: disable=W0622
    return_value = next(root, info, **args)
    return return_value


def error_middleware(next, root, info, **args):  #pylint: disable=W0622
    raise RuntimeError("Runtime Error!")


_runtime_error_name = callable_name(RuntimeError)
_test_runtime_error = [(_runtime_error_name, "Runtime Error!")]
_graphql_base_rollup_metrics = [
    ("OtherTransaction/all", 1),
    ("GraphQL/all", 1),
    ("GraphQL/allOther", 1),
    ("GraphQL/Strawberry/all", 1),
    ("GraphQL/Strawberry/allOther", 1),
]


def test_basic(app, graphql_run):
    from graphql import __version__ as version

    from newrelic.hooks.framework_strawberry import framework_details

    FRAMEWORK_METRICS = [
        ("Python/Framework/Strawberry/%s" % framework_details()[1], 1),
        ("Python/Framework/GraphQL/%s" % version, 1),
    ]

    @validate_transaction_metrics(
        "query/<anonymous>/hello",
        "GraphQL",
        rollup_metrics=_graphql_base_rollup_metrics + FRAMEWORK_METRICS,
        background_task=True,
    )
    @background_task()
    def _test():
        response = graphql_run(app, "{ hello }")
        assert not response.errors

    _test()


@dt_enabled
def test_query_and_mutation(app, graphql_run):
    from graphql import __version__ as version

    from newrelic.hooks.framework_strawberry import framework_details

    FRAMEWORK_METRICS = [
        ("Python/Framework/Strawberry/%s" % framework_details()[1], 1),
        ("Python/Framework/GraphQL/%s" % version, 1),
    ]
    _test_mutation_scoped_metrics = [
        ("GraphQL/resolve/Strawberry/storage", 1),
        ("GraphQL/resolve/Strawberry/storage_add", 1),
        ("GraphQL/operation/Strawberry/query/<anonymous>/storage", 1),
        ("GraphQL/operation/Strawberry/mutation/<anonymous>/storage_add", 1),
    ]
    _test_mutation_unscoped_metrics = [
        ("OtherTransaction/all", 1),
        ("GraphQL/all", 2),
        ("GraphQL/Strawberry/all", 2),
        ("GraphQL/allOther", 2),
        ("GraphQL/Strawberry/allOther", 2),
    ] + _test_mutation_scoped_metrics

    _expected_mutation_operation_attributes = {
        "graphql.operation.type": "mutation",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_mutation_resolver_attributes = {
        "graphql.field.name": "storage_add",
        "graphql.field.parentType": "Mutation",
        "graphql.field.path": "storage_add",
        "graphql.field.returnType": "String!",
    }
    _expected_query_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_query_resolver_attributes = {
        "graphql.field.name": "storage",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "storage",
        "graphql.field.returnType": "[String!]!",
    }

    @validate_transaction_metrics(
        "query/<anonymous>/storage",
        "GraphQL",
        scoped_metrics=_test_mutation_scoped_metrics,
        rollup_metrics=_test_mutation_unscoped_metrics + FRAMEWORK_METRICS,
        background_task=True,
    )
    @validate_span_events(exact_agents=_expected_mutation_operation_attributes)
    @validate_span_events(exact_agents=_expected_mutation_resolver_attributes)
    @validate_span_events(exact_agents=_expected_query_operation_attributes)
    @validate_span_events(exact_agents=_expected_query_resolver_attributes)
    @background_task()
    def _test():
        response = graphql_run(app, 'mutation { storage_add(string: "abc") }')
        assert not response.errors
        response = graphql_run(app, "query { storage }")
        assert not response.errors

        # These are separate assertions because pypy stores 'abc' as a unicode string while other Python versions do not
        assert "storage" in str(response.data)
        assert "abc" in str(response.data)

    _test()


@pytest.mark.parametrize("field", ("error", "error_non_null"))
@dt_enabled
def test_exception_in_resolver(app, graphql_run, field):
    query = "query MyQuery { %s }" % field

    txn_name = "_target_application:resolve_error"

    # Metrics
    _test_exception_scoped_metrics = [
        ("GraphQL/operation/Strawberry/query/MyQuery/%s" % field, 1),
        ("GraphQL/resolve/Strawberry/%s" % field, 1),
    ]
    _test_exception_rollup_metrics = [
        ("Errors/all", 1),
        ("Errors/allOther", 1),
        ("Errors/OtherTransaction/GraphQL/%s" % txn_name, 1),
    ] + _test_exception_scoped_metrics

    # Attributes
    _expected_exception_resolver_attributes = {
        "graphql.field.name": field,
        "graphql.field.parentType": "Query",
        "graphql.field.path": field,
        "graphql.field.returnType": "String!" if "non_null" in field else "String",
    }
    _expected_exception_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "MyQuery",
        "graphql.operation.query": query,
    }

    @validate_transaction_metrics(
        txn_name,
        "GraphQL",
        scoped_metrics=_test_exception_scoped_metrics,
        rollup_metrics=_test_exception_rollup_metrics + _graphql_base_rollup_metrics,
        background_task=True,
    )
    @validate_span_events(exact_agents=_expected_exception_operation_attributes)
    @validate_span_events(exact_agents=_expected_exception_resolver_attributes)
    @validate_transaction_errors(errors=_test_runtime_error)
    @background_task()
    def _test():
        response = graphql_run(app, query)
        assert response.errors

    _test()


@dt_enabled
@pytest.mark.parametrize(
    "query,exc_class",
    [
        ("query MyQuery { missing_field }", "GraphQLError"),
        ("{ syntax_error ", "graphql.error.syntax_error:GraphQLSyntaxError"),
    ],
)
def test_exception_in_validation(app, graphql_run, is_graphql_2, query, exc_class):
    if "syntax" in query:
        txn_name = "graphql.language.parser:parse"
    else:
        if is_graphql_2:
            txn_name = "graphql.validation.validation:validate"
        else:
            txn_name = "graphql.validation.validate:validate"

    # Import path differs between versions
    if exc_class == "GraphQLError":
        from graphql.error import GraphQLError

        exc_class = callable_name(GraphQLError)

    _test_exception_scoped_metrics = [
            ('GraphQL/operation/Strawberry/<unknown>/<anonymous>/<unknown>', 1),
    ]
    _test_exception_rollup_metrics = [
        ("Errors/all", 1),
        ("Errors/allOther", 1),
        ("Errors/OtherTransaction/GraphQL/%s" % txn_name, 1),
    ] + _test_exception_scoped_metrics

    # Attributes
    _expected_exception_operation_attributes = {
        "graphql.operation.type": "<unknown>",
        "graphql.operation.name": "<anonymous>",
        "graphql.operation.query": query,
    }

    @validate_transaction_metrics(
        txn_name,
        "GraphQL",
        scoped_metrics=_test_exception_scoped_metrics,
        rollup_metrics=_test_exception_rollup_metrics + _graphql_base_rollup_metrics,
        background_task=True,
    )
    @validate_span_events(exact_agents=_expected_exception_operation_attributes)
    @validate_transaction_errors(errors=[exc_class])
    @background_task()
    def _test():
        response = graphql_run(app, query)
        assert response.errors

    _test()


@dt_enabled
def test_operation_metrics_and_attrs(app, graphql_run):
    operation_metrics = [("GraphQL/operation/Strawberry/query/MyQuery/library", 1)]
    operation_attrs = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "MyQuery",
    }

    @validate_transaction_metrics(
        "query/MyQuery/library",
        "GraphQL",
        scoped_metrics=operation_metrics,
        rollup_metrics=operation_metrics + _graphql_base_rollup_metrics,
        background_task=True,
    )
    # Span count 16: Transaction, Operation, and 7 Resolvers and Resolver functions
    # library, library.name, library.book
    # library.book.name and library.book.id for each book resolved (in this case 2)
    @validate_span_events(count=16)
    @validate_span_events(exact_agents=operation_attrs)
    @background_task()
    def _test():
        response = graphql_run(app, "query MyQuery { library(index: 0) { branch, book { id, name } } }")
        assert not response.errors

    _test()


@dt_enabled
def test_field_resolver_metrics_and_attrs(app, graphql_run):
    field_resolver_metrics = [("GraphQL/resolve/Strawberry/hello", 1)]
    graphql_attrs = {
        "graphql.field.name": "hello",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "hello",
        "graphql.field.returnType": "String!",
    }

    @validate_transaction_metrics(
        "query/<anonymous>/hello",
        "GraphQL",
        scoped_metrics=field_resolver_metrics,
        rollup_metrics=field_resolver_metrics + _graphql_base_rollup_metrics,
        background_task=True,
    )
    # Span count 4: Transaction, Operation, and 1 Resolver and Resolver function
    @validate_span_events(count=4)
    @validate_span_events(exact_agents=graphql_attrs)
    @background_task()
    def _test():
        response = graphql_run(app, "{ hello }")
        assert not response.errors
        assert "Hello!" in str(response.data)

    _test()


_test_queries = [
    ("{ hello }", "{ hello }"),  # Basic query extraction
    ("{ error }", "{ error }"),  # Extract query on field error
    (to_graphql_source("{ hello }"), "{ hello }"),  # Extract query from Source objects
    (
        "{ library(index: 0) { branch } }",
        "{ library(index: ?) { branch } }",
    ),  # Integers
    ('{ echo(echo: "123") }', "{ echo(echo: ?) }"),  # Strings with numerics
    ('{ echo(echo: "test") }', "{ echo(echo: ?) }"),  # Strings
    ('{ TestEcho: echo(echo: "test") }', "{ TestEcho: echo(echo: ?) }"),  # Aliases
    ('{ TestEcho: echo(echo: "test") }', "{ TestEcho: echo(echo: ?) }"),  # Variables
    (  # Fragments
        '{ ...MyFragment } fragment MyFragment on Query { echo(echo: "test") }',
        "{ ...MyFragment } fragment MyFragment on Query { echo(echo: ?) }",
    ),
]


@dt_enabled
@pytest.mark.parametrize("query,obfuscated", _test_queries)
def test_query_obfuscation(app, graphql_run, query, obfuscated):
    graphql_attrs = {"graphql.operation.query": obfuscated}

    if callable(query):
        query = query()

    @validate_span_events(exact_agents=graphql_attrs)
    @background_task()
    def _test():
        response = graphql_run(app, query)
        if not isinstance(query, str) or "error" not in query:
            assert not response.errors

    _test()


_test_queries = [
    ("{ hello }", "/hello"),  # Basic query
    ("{ error }", "/error"),  # Extract deepest path on field error
    ('{ echo(echo: "test") }', "/echo"),  # Fields with arguments
    (
        "{ library(index: 0) { branch, book { isbn branch } } }",
        "/library",
    ),  # Complex Example, 1 level
    (
        "{ library(index: 0) { book { author { first_name }} } }",
        "/library.book.author.first_name",
    ),  # Complex Example, 2 levels
    ("{ library(index: 0) { id, book { name } } }", "/library.book.name"),  # Filtering
    ('{ TestEcho: echo(echo: "test") }', "/echo"),  # Aliases
    (
        '{ search(contains: "A") { __typename ... on Book { name } } }',
        "/search<Book>.name",
    ),  # InlineFragment
    (
        '{ hello echo(echo: "test") }',
        "",
    ),  # Multiple root selections. (need to decide on final behavior)
    # FragmentSpread
    (
        "{ library(index: 0) { book { ...MyFragment } } } fragment MyFragment on Book { name id }",  # Fragment filtering
        "/library.book.name",
    ),
    (
        "{ library(index: 0) { book { ...MyFragment } } } fragment MyFragment on Book { author { first_name } }",
        "/library.book.author.first_name",
    ),
    (
        "{ library(index: 0) { book { ...MyFragment } magazine { ...MagFragment } } } fragment MyFragment on Book { author { first_name } } fragment MagFragment on Magazine { name }",
        "/library",
    ),
]


@dt_enabled
@pytest.mark.parametrize("query,expected_path", _test_queries)
def test_deepest_unique_path(app, graphql_run, query, expected_path):
    if expected_path == "/error":
        txn_name = "_target_application:resolve_error"
    else:
        txn_name = "query/<anonymous>%s" % expected_path

    @validate_transaction_metrics(
        txn_name,
        "GraphQL",
        background_task=True,
    )
    @background_task()
    def _test():
        response = graphql_run(app, query)
        if "error" not in query:
            assert not response.errors

    _test()


@validate_transaction_count(0)
@background_task()
def test_ignored_introspection_transactions(app, graphql_run):
    response = graphql_run(app, "{ __schema { types { name } } }")
    assert not response.errors
